from os import path

import pytest

import autofit as af
from autofit.mock.mock_model import MockClassx4

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


class TestEmceeConfig:

    def test__loads_from_config_file_correct(self):

        emcee = af.Emcee(
            prior_passer=af.PriorPasser(sigma=2.0, use_errors=False, use_widths=False),
            nwalkers=51,
            nsteps=2001,
            initializer=af.InitializerBall(lower_limit=0.2, upper_limit=0.8),
            auto_correlations_settings=af.AutoCorrelationsSettings(
                check_for_convergence=False,
                check_size=101,
                required_length=51,
                change_threshold=0.02
            ),
            number_of_cores=2,
        )

        assert emcee.prior_passer.sigma == 2.0
        assert emcee.prior_passer.use_errors is False
        assert emcee.prior_passer.use_widths is False
        assert emcee.config_dict_search["nwalkers"] == 51
        assert emcee.config_dict_run["nsteps"] == 2001
        assert isinstance(emcee.initializer, af.InitializerBall)
        assert emcee.initializer.lower_limit == 0.2
        assert emcee.initializer.upper_limit == 0.8
        assert emcee.auto_correlations_settings.check_for_convergence is False
        assert emcee.auto_correlations_settings.check_size == 101
        assert emcee.auto_correlations_settings.required_length == 51
        assert emcee.auto_correlations_settings.change_threshold == 0.02
        assert emcee.number_of_cores == 2

        emcee = af.Emcee()

        assert emcee.prior_passer.sigma == 3.0
        assert emcee.prior_passer.use_errors is True
        assert emcee.prior_passer.use_widths is True
        assert emcee.config_dict_search["nwalkers"] == 50
        assert emcee.config_dict_run["nsteps"] == 2000
        assert isinstance(emcee.initializer, af.InitializerPrior)
        assert emcee.auto_correlations_settings.check_for_convergence is True
        assert emcee.auto_correlations_settings.check_size == 100
        assert emcee.auto_correlations_settings.required_length == 50
        assert emcee.auto_correlations_settings.change_threshold == 0.01
        assert emcee.number_of_cores == 1

    def test__samples_from_model(self):

        emcee = af.Emcee()
        emcee.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "emcee"))
        emcee.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        samples = emcee.samples_from(model=model)

        assert isinstance(samples.parameter_lists, list)
        assert isinstance(samples.parameter_lists[0], list)
        assert isinstance(samples.log_likelihood_list, list)
        assert isinstance(samples.log_prior_list, list)
        assert isinstance(samples.log_posterior_list, list)
        assert isinstance(samples.weight_list, list)

        assert samples.parameter_lists[0] == pytest.approx(
            [0.173670, 0.162607, 3095.28, 0.62104], 1.0e-4
        )
        assert samples.log_likelihood_list[0] == pytest.approx(-17257775239.32677, 1.0e-4)
        assert samples.log_prior_list[0] == pytest.approx(1.6102016075510708, 1.0e-4)
        assert samples.weight_list[0] == pytest.approx(1.0, 1.0e-4)
        assert samples.total_steps == 1000
        assert samples.total_walkers == 10
        assert samples.auto_correlations.times[0] == pytest.approx(31.98507, 1.0e-4)


class TestEmceeOutput:
    def test__median_pdf_parameters(self):
        emcee = af.Emcee()
        emcee.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "emcee"))
        emcee.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        samples = emcee.samples_from(model=model)

        assert samples.median_pdf_vector == pytest.approx(
            [0.008422, -0.026413, 9.9579656, 0.494618], 1.0e-3
        )

    def test__vector_at_sigma__uses_output_files(self):
        emcee = af.Emcee()
        emcee.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "emcee"))
        emcee.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        samples = emcee.samples_from(model=model)

        parameters = samples.vector_at_sigma(sigma=3.0)

        assert parameters[0][0:2] == pytest.approx((-0.003197, 0.019923), 1e-2)

        parameters = samples.vector_at_sigma(sigma=1.0)

        assert parameters[0][0:2] == pytest.approx((0.0042278, 0.01087681), 1e-2)

    def test__autocorrelation_times(self):
        emcee = af.Emcee()
        emcee.paths = af.DirectoryPaths(path_prefix=path.join("non_linear", "emcee"))
        emcee.paths._identifier = "tag"

        model = af.ModelMapper(mock_class=MockClassx4)
        model.mock_class.two = af.LogUniformPrior(lower_limit=1e-8, upper_limit=10.0)

        samples = emcee.samples_from(model=model)

        assert samples.auto_correlations.previous_times == pytest.approx(
            [31.1079, 36.0910, 72.44768, 65.86194], 1.0e-4
        )
        assert samples.auto_correlations.times == pytest.approx(
            [31.98507, 36.51001, 73.47629, 67.67495], 1.0e-4
        )



from typing import Optional

import autofit as af


class OptionalClass:
    def __init__(self, optional: Optional[af.Gaussian]):
        self.optional = optional


def test_optional():
    model = af.Model(
        OptionalClass
    )

    assert model.instance_from_prior_medians().optional is None


class DefaultClass:
    def __init__(self, default=1.0):
        self.default = default

    __default_fields__ = ("default",)


def test_default():
    model = af.Model(
        DefaultClass
    )
    assert model.instance_from_prior_medians().default == 1.0




import autofit as af
from autofit.database import query as q

def test_simple(aggregator):
    assert aggregator.model.centre.query == q.Q("centre").query


def test_and(aggregator):
    construction = ((aggregator.model.centre == af.Gaussian) & (aggregator.model.centre.x == 0))
    assert construction.query == q.Q(
        "centre",
        q.And(
            q.T(af.Gaussian),
            q.Q(
                "x",
                q.V(
                    "=", 0
                )
            )
        )
    ).query


def test_or(aggregator):
    construction = ((aggregator.model.centre == af.Gaussian) | (aggregator.model.centre.x == 0))
    assert construction.query == q.Q(
        "centre",
        q.Or(
            q.T(af.Gaussian),
            q.Q(
                "x",
                q.V(
                    "=", 0
                )
            )
        )
    ).query


def test_with_value(aggregator):
    assert (aggregator.model.centre == 1).query == q.Q("centre", q.V("=", 1)).query


def test_second_level(aggregator):
    assert (aggregator.model.lens.centre == 1).query == q.Q("lens", q.Q("centre", q.V("=", 1))).query


def test_third_level(aggregator):
    assert (aggregator.model.lens.centre.x == 1).query == q.Q("lens", q.Q("centre", q.Q("x", q.V("=", 1)))).query


def test_with_type(aggregator):
    assert (aggregator.model.centre == af.Gaussian).query == q.Q("centre", q.T(af.Gaussian)).query


def test_with_string(aggregator):
    assert (aggregator.model.centre == "centre").query == q.Q("centre", q.SV("=", "centre")).query


def test_greater_than(aggregator):
    assert (aggregator.model.centre > 1).query == q.Q("centre", q.V(">", 1)).query


def test_less_than(aggregator):
    assert (aggregator.model.centre < 1).query == q.Q("centre", q.V("<", 1)).query


def test_less_than_equals(aggregator):
    assert (aggregator.model.centre <= 1).query == q.Q("centre", q.V("<=", 1)).query


def test_greater_than_equals(aggregator):
    assert (aggregator.model.centre >= 1).query == q.Q("centre", q.V(">=", 1)).query



import os
import pytest

import autofit as af
from autofit.mock.mock_model import MockClassx2, MockClassx4
from autofit.mock.mock import MockSamples

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


def test__table__headers(samples_x5):
    assert samples_x5._headers == [
        "mock_class_1_one",
        "mock_class_1_two",
        "mock_class_1_three",
        "mock_class_1_four",
        "log_likelihood",
        "log_prior",
        "log_posterior",
        "weight",
    ]


def test__table__rows(samples_x5):
    rows = list(samples_x5._rows)
    assert rows == [
        [0.0, 1.0, 2.0, 3.0, 1.0, 0.0, 1.0, 1.0],
        [0.0, 1.0, 2.0, 3.0, 2.0, 0.0, 2.0, 1.0],
        [0.0, 1.0, 2.0, 3.0, 3.0, 0.0, 3.0, 1.0],
        [21.0, 22.0, 23.0, 24.0, 10.0, 0.0, 10.0, 1.0],
        [0.0, 1.0, 2.0, 3.0, 5.0, 0.0, 5.0, 1.0],
    ]


def test__table__write_table():

    model = af.Collection(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    samples_x5 = af.Samples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )

    filename = "samples.csv"
    samples_x5.write_table(filename=filename)

    assert os.path.exists(filename)
    os.remove(filename)


def test__max_log_likelihood_vector_and_instance(samples_x5):
    assert samples_x5.max_log_likelihood_vector == [21.0, 22.0, 23.0, 24.0]

    instance = samples_x5.max_log_likelihood_instance

    assert instance.mock_class_1.one == 21.0
    assert instance.mock_class_1.two == 22.0
    assert instance.mock_class_1.three == 23.0
    assert instance.mock_class_1.four == 24.0


def test__log_prior_list_and_max_log_posterior_vector_and_instance():
    model = af.Collection(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
    ]

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 0.0, 5.0],
            log_prior_list=[1.0, 2.0, 3.0, 10.0, 6.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )

    assert samples_x5.log_posterior_list == [2.0, 4.0, 6.0, 10.0, 11.0]

    assert samples_x5.max_log_posterior_vector == [21.0, 22.0, 23.0, 24.0]

    instance = samples_x5.max_log_posterior_instance

    assert instance.mock_class_1.one == 21.0
    assert instance.mock_class_1.two == 22.0
    assert instance.mock_class_1.three == 23.0
    assert instance.mock_class_1.four == 24.0


def test__gaussian_priors():
    parameters = [
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.1],
        [1.0, 2.0, 3.0, 4.1],
        [0.88, 1.88, 2.88, 3.88],
        [1.12, 2.12, 3.12, 4.32],
    ]

    model = af.Collection(mock_class=MockClassx4)
    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[10.0, 0.0, 0.0, 0.0, 0.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )

    gaussian_priors = samples_x5.gaussian_priors_at_sigma(sigma=1.0)

    assert gaussian_priors[0][0] == 1.0
    assert gaussian_priors[1][0] == 2.0
    assert gaussian_priors[2][0] == 3.0
    assert gaussian_priors[3][0] == 4.0

    assert gaussian_priors[0][1] == pytest.approx(0.12, 1.0e-4)
    assert gaussian_priors[1][1] == pytest.approx(0.12, 1.0e-4)
    assert gaussian_priors[2][1] == pytest.approx(0.12, 1.0e-4)
    assert gaussian_priors[3][1] == pytest.approx(0.32, 1.0e-4)


def test__instance_from_sample_index():
    model = af.Collection(mock_class=MockClassx4)

    parameters = [
        [1.0, 2.0, 3.0, 4.0],
        [5.0, 6.0, 7.0, 8.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.0, 2.0, 3.0, 4.0],
        [1.1, 2.1, 3.1, 4.1],
    ]

    samples_x5 = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )

    instance = samples_x5.instance_from_sample_index(sample_index=0)

    assert instance.mock_class.one == 1.0
    assert instance.mock_class.two == 2.0
    assert instance.mock_class.three == 3.0
    assert instance.mock_class.four == 4.0

    instance = samples_x5.instance_from_sample_index(sample_index=1)

    assert instance.mock_class.one == 5.0
    assert instance.mock_class.two == 6.0
    assert instance.mock_class.three == 7.0
    assert instance.mock_class.four == 8.0


def test__addition_of_samples(samples_x5):

    samples = samples_x5 + samples_x5

    assert len(samples.sample_list) == 10
    assert samples.sample_list[0].log_likelihood == 1.0
    assert samples.sample_list[4].log_likelihood == 5.0
    assert samples.sample_list[5].log_likelihood == 1.0
    assert samples.sample_list[9].log_likelihood == 5.0

def test__sum_of_samples(samples_x5):

    samples = sum([samples_x5, samples_x5, samples_x5])

    assert len(samples.sample_list) == 15
    assert samples.sample_list[0].log_likelihood == 1.0
    assert samples.sample_list[4].log_likelihood == 5.0
    assert samples.sample_list[5].log_likelihood == 1.0
    assert samples.sample_list[9].log_likelihood == 5.0
    assert samples.sample_list[10].log_likelihood == 1.0
    assert samples.sample_list[14].log_likelihood == 5.0

def test__addition_of_samples__raises_error_if_model_mismatch(samples_x5):

    model = af.Collection(mock_class_1=MockClassx2)

    parameters = [
        [0.0, 1.0],
        [0.0, 1.0],
        [0.0, 1.0],
        [21.0, 22.0],
        [0.0, 1.0],
    ]

    samples_different_model = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0],
            log_prior_list=[0.0, 0.0],
            weight_list=[1.0, 1.0],
        ),
    )

    with pytest.raises(af.exc.SamplesException):
        samples_x5 + samples_different_model


import pytest

import autofit as af


@pytest.fixture(
    name="model"
)
def make_model():
    model = af.Model(
        af.Gaussian
    )
    return model


def test_sort_alphabetically(
        model
):
    assert model.sort_priors_alphabetically([
        model.sigma,
        model.centre,
        model.normalization,
    ]) == [
               model.centre,
               model.normalization,
               model.sigma,
           ]



import pytest

import autofit as af


class Analysis(af.Analysis):

    def log_likelihood_function(self, instance):
        return -1


@pytest.fixture(
    name="search"
)
def make_search(session):
    return af.DynestyStatic(
        session=session
    )


@pytest.fixture(
    name="model"
)
def make_model():
    return af.Model(
        af.Gaussian
    )


def test_is_database_paths(search):
    assert isinstance(
        search.paths,
        af.DatabasePaths
    )


# def test_is_complete(search, session, model):
#     search.fit(
#         model,
#         Analysis()
#     )
#
#     fit, = m.Fit.all(session)
#
#     assert fit.id == search.paths.identifier
#     assert fit.is_complete
#
#     search = af.DynestyStatic(
#         session=session
#     )
#
#     search.paths.model = model
#
#     assert search.paths.identifier == fit.id
#     assert search.paths.is_complete
#
#     search.fit(
#         model,
#         Analysis()
#     )
#
#
# def test_remove_after(search, model):
#     search.paths.model = model
#     output_path = search.paths.output_path
#
#     search.fit(
#         model,
#         Analysis()
#     )
#
#     assert not os.path.exists(
#         output_path
#     )



import autofit as af
from autofit import database as db


def test_set_and_retrieve(
        session
):
    fit = db.Fit()
    fit.named_instances[
        "one"
    ] = af.Gaussian()

    assert isinstance(
        fit.named_instances[
            "one"
        ],
        af.Gaussian
    )


def test_query(
        session
):
    fit = db.Fit(
        id="test_query"
    )
    fit.named_instances[
        "one"
    ] = af.Gaussian()

    session.add(fit)
    session.commit()

    fit, = db.Fit.all(
        session
    )

    assert isinstance(
        fit.named_instances[
            "one"
        ],
        af.Gaussian
    )



import autofit as af
from autofit import graphical as g
from autofit.mock.mock import MockAnalysis


class Analysis(MockAnalysis):
    def __init__(self):
        super().__init__()
        self.did_call_visualise = False

    def visualize(self, paths, instance, during_analysis):
        self.did_call_visualise = True


def test_visualize():
    analysis = Analysis()

    gaussian = af.Model(af.Gaussian)

    analysis_factor = g.AnalysisFactor(
        prior_model=gaussian,
        analysis=analysis
    )

    factor_graph = g.FactorGraphModel(
        analysis_factor
    )

    model = factor_graph.global_prior_model
    instance = model.instance_from_prior_medians()

    factor_graph.visualize(
        af.DirectoryPaths(),
        instance,
        False
    )

    assert analysis.did_call_visualise is True



from autoconf.conf import with_config
from autofit import graphical as g, DirectoryPaths

from autofit.mock import mock as m

MAX_STEPS = 3


class MockResult(m.MockResult):
    def __init__(self, model):
        super().__init__()
        self.model = model

    @property
    def projected_model(self):
        return self.model


class MockSearch(m.MockSearch):
    def fit(
            self,
            model,
            analysis,
            info=None,
            pickle_files=None,
            log_likelihood_cap=None,
    ):
        super().fit(model, analysis)
        return MockResult(model)


def _run_optimisation(factor_graph_model, paths=None):
    search = MockSearch()
    factor_graph_model.optimise(
        search,
        max_steps=MAX_STEPS,
        log_interval=1,
        visualise_interval=1,
        output_interval=1,
        paths=paths
              or DirectoryPaths(
            name="name",
            is_identifier_in_paths=False,
        ),
    )


@with_config("general", "output", "remove_files", value=False)
def test_output(output_directory, factor_graph_model):
    factor_graph_model.model_factors[0]._name = "factor_1"
    factor_graph_model.model_factors[1]._name = "factor_2"
    _run_optimisation(factor_graph_model)

    path = output_directory / "name/factor_1"

    assert path.exists()
    assert (output_directory / "name/factor_2").exists()

    for number in range(MAX_STEPS):
        assert (path / f"optimization_{number}").exists()


@with_config("general", "output", "remove_files", value=False)
def test_default_output(output_directory, factor_graph_model):
    _run_optimisation(factor_graph_model)
    assert (output_directory / "name/AnalysisFactor0").exists()
    assert (output_directory / "name/AnalysisFactor1").exists()


@with_config("general", "output", "remove_files", value=False)
def test_path_prefix(output_directory, factor_graph_model):
    paths = DirectoryPaths(
        path_prefix="path_prefix",
        name="name",
        is_identifier_in_paths=False,
    )
    optimiser = g.EPOptimiser(
        factor_graph=factor_graph_model.graph,
        paths=paths,
        default_optimiser=MockSearch(),
    )

    assert optimiser.output_path == output_directory / "path_prefix/name"

    _run_optimisation(factor_graph_model, paths=paths)
    assert (output_directory / "path_prefix/name/AnalysisFactor0").exists()
    assert (output_directory / "path_prefix/name/AnalysisFactor1").exists()



import pytest

import autofit as af
from autofit import graphical as g
from autofit.graphical.declarative.graph import GraphInfoFormatter
from autofit.mock.mock import MockAnalysis


@pytest.fixture(name="model_factor_1")
def make_model_factor_1():
    return g.AnalysisFactor(af.Model(af.Gaussian), MockAnalysis())


@pytest.fixture(name="model_factor_2")
def make_model_factor_2():
    return g.AnalysisFactor(af.Model(af.Gaussian), MockAnalysis())


@pytest.fixture(name="hierarchical_factor")
def make_hierarchical_factor(
        model_factor_1,
        model_factor_2,
):
    hierarchical_factor = g.HierarchicalFactor(
        af.GaussianPrior,
        mean=af.GaussianPrior(mean=100, sigma=10),
        sigma=af.GaussianPrior(mean=10, sigma=5),
    )

    hierarchical_factor.add_drawn_variable(model_factor_1.centre)
    hierarchical_factor.add_drawn_variable(model_factor_2.centre)
    return hierarchical_factor


@pytest.fixture(name="hierarchical_model")
def make_factor_graph_model(
        model_factor_1,
        model_factor_2,
        hierarchical_factor,
):
    return g.FactorGraphModel(model_factor_1, model_factor_2, hierarchical_factor)


def test_hierarchical_factors(graph, hierarchical_factor):
    assert graph.hierarchical_factors == [hierarchical_factor]


@pytest.fixture(name="graph")
def make_graph(hierarchical_model):
    return hierarchical_model.graph


def test_info_for_hierarchical_factor(graph):
    info = GraphInfoFormatter(graph).info_for_hierarchical_factor(
        graph.hierarchical_factors[0]
    )
    assert (
            info
            == """HierarchicalFactor0

mean (HierarchicalFactor0, PriorFactor1)                                                  GaussianPrior, mean = 100, sigma = 10
sigma (HierarchicalFactor0, PriorFactor0)                                                 GaussianPrior, mean = 10, sigma = 5

Drawn Variables

AnalysisFactor0.centre, PriorFactor7                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
AnalysisFactor1.centre, PriorFactor4                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0"""
    )


def test_graph_info(graph):
    info = graph.info
    assert (
            info
            == """PriorFactors

PriorFactor0 (HierarchicalFactor0)                                                        GaussianPrior, mean = 10, sigma = 5
PriorFactor1 (HierarchicalFactor0)                                                        GaussianPrior, mean = 100, sigma = 10
PriorFactor2 (AnalysisFactor1.sigma)                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor3 (AnalysisFactor1.normalization)                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor4 (AnalysisFactor1.centre, HierarchicalFactor0)                                UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor5 (AnalysisFactor0.sigma)                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor6 (AnalysisFactor0.normalization)                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
PriorFactor7 (AnalysisFactor0.centre, HierarchicalFactor0)                                UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactors

AnalysisFactor0

centre (HierarchicalFactor0, PriorFactor7)                                                UniformPrior, lower_limit = 0.0, upper_limit = 1.0
normalization (PriorFactor6)                                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
sigma (PriorFactor5)                                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0

AnalysisFactor1

centre (HierarchicalFactor0, PriorFactor4)                                                UniformPrior, lower_limit = 0.0, upper_limit = 1.0
normalization (PriorFactor3)                                                              UniformPrior, lower_limit = 0.0, upper_limit = 1.0
sigma (PriorFactor2)                                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0

HierarchicalFactors

HierarchicalFactor0

mean (HierarchicalFactor0, PriorFactor1)                                                  GaussianPrior, mean = 100, sigma = 10
sigma (HierarchicalFactor0, PriorFactor0)                                                 GaussianPrior, mean = 10, sigma = 5

Drawn Variables

AnalysisFactor0.centre, PriorFactor7                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0
AnalysisFactor1.centre, PriorFactor4                                                      UniformPrior, lower_limit = 0.0, upper_limit = 1.0"""
    )



import logging
from typing import Dict, Tuple, Optional, List

import numpy as np

from autoconf import cached_property
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.mean_field import MeanField, FactorApproximation
from autofit.graphical.utils import Status
from autofit.mapper.variable import Variable
from autofit.messages.abstract import AbstractMessage

logger = logging.getLogger(__name__)


class EPMeanField(FactorGraph):
    """
    this class encode the EP mean-field approximation to a factor graph


    Attributes
    ----------
    factor_graph: FactorGraph
        the base factor graph being approximated

    factor_mean_field: Dict[Factor, MeanField]
        the mean-field approximation for each factor in the factor graph

    mean_field: MeanField
        the mean-field approximation of the full factor graph
        i.e. the product of the factor mean-field approximations

    variables: Set[Variable]
        the variables of the approximation

    deterministic_variables: Set[Variable]
        the deterministic variables

    log_evidence: float
        the approximate log evidence of the approximation

    is_valid: bool
        returns whether the factor mean-field approximations are all valid

    Methods
    -------
    from_approx_dists(factor_graph, approx_dists)
        create a EPMeanField object from the passed factor_graph
        using approx_dists to initialise the factor mean-field approximations

    factor_approximation(factor)
        create the FactorApproximation for the factor

    project_factor_approx(factor_approximation)
        given the passed FactorApproximation, return a new `EPMeanField`
        object encoding the updated mean-field approximation
    """

    def __init__(
            self, factor_graph: FactorGraph, factor_mean_field: Dict[Factor, MeanField]
    ):
        self._factor_graph = factor_graph
        self._factor_mean_field = factor_mean_field

        super().__init__(self.factor_graph.factors)

    @property
    def name(self):
        return f"EP_{self.factor_graph.name}"

    @cached_property
    def fixed_values(self):
        return {
            k: v for mf in self._factor_mean_field.values() 
            for k, v in mf.fixed_values.items()
        }

    @property
    def variables(self):
        return self.factor_graph.variables

    @property
    def deterministic_variables(self):
        return self.factor_graph.deterministic_variables

    @property
    def factor_mean_field(self) -> Dict[Factor, MeanField]:
        return self._factor_mean_field.copy()

    @property
    def factor_graph(self) -> FactorGraph:
        return self._factor_graph

    @classmethod
    def from_approx_dists(
            cls,
            factor_graph: FactorGraph,
            approx_dists: Dict[Variable, AbstractMessage],
    ) -> "EPMeanField":
        factor_mean_field = {
            factor: MeanField({v: approx_dists[v] for v in factor.all_variables})
            for factor in factor_graph.factors
        }

        return cls(factor_graph, factor_mean_field)

    from_kws = from_approx_dists

    def factor_approximation(self, factor: Factor) -> FactorApproximation:
        """
        Create an approximation for one factor.

        This comprises:
        - The factor
        - The factor's variable distributions
        - The cavity distribution, which is the product of the distributions
        for each variable for all other factors
        - The model distribution, which is the product of the distributions
        for each variable for all factors

        Parameters
        ----------
        factor
            Some factor

        Returns
        -------
        An object comprising distributions with a specific distribution excluding
        that factor
        """
        factor_mean_field = self._factor_mean_field.copy()
        factor_dist = factor_mean_field.pop(factor)
        cavity_dist = MeanField({v: 1.0 for v in factor_dist.all_variables}).prod(
            *factor_mean_field.values()
        )

        model_dist = factor_dist.prod(cavity_dist)

        return FactorApproximation(factor, cavity_dist, factor_dist, model_dist)

    def project_factor_approx(
            self,
            projection: FactorApproximation,
            status: Optional[Status] = None,
    ) -> Tuple["EPMeanField", Status]:
        """ """
        factor_mean_field = self.factor_mean_field
        factor_mean_field[projection.factor] = projection.factor_dist

        new_approx = type(self)(
            factor_graph=self._factor_graph, factor_mean_field=factor_mean_field
        )
        return new_approx, status

    project = project_factor_approx

    @property
    def mean_field(self) -> MeanField:
        return MeanField({v: 1.0 for v in self.all_variables}).prod(
            *self._factor_mean_field.values()
        )

    model_dist = mean_field

    @property
    def variable_messages(self) -> Dict[Variable, List[AbstractMessage]]:
        variable_messages = {v: [] for v in self.all_variables}
        for meanfield in self.factor_mean_field.values():
            for v, message in meanfield.items():
                variable_messages[v].append(message)

        return variable_messages

    @property
    def variable_evidence(self) -> Dict[Variable, np.ndarray]:
        return {
            v: AbstractMessage.log_normalisation(*ms)
            for v, ms in self.variable_messages.items()
        }

    @property
    def factor_evidence(self) -> Dict[Factor, np.ndarray]:
        return {
            factor: meanfield.log_norm
            for factor, meanfield in self.factor_mean_field.items()
        }

    @property
    def log_evidence(self):
        """
        Calculates evidence for the EP approximation

        Evidence for a variable, xᵢ,

        Zᵢ = ∫ ∏ₐ m_{a → i} (xᵢ) dxᵢ

        Evidence for a factor, f_a,

                ∫ ∏_{j ∈ a} m_{i → a} (xᵢ) fₐ(xₐ) dxₐ
        Zₐ = -----------------------------------------
                             ∏_{j ∈ a} Zⱼ

        Evidence for model

        Z = ∏ᵢ Zᵢ ∏ₐ Zₐ
        """
        variable_evidence = {
            v: np.sum(logz) for v, logz in self.variable_evidence.items()
        }
        factor_evidence = sum(
            np.sum(meanfield.log_norm)
            - sum(variable_evidence[v] for v in factor.all_variables)
            for factor, meanfield in self.factor_mean_field.items()
        )
        return factor_evidence + sum(variable_evidence.values())

    def __repr__(self) -> str:
        clsname = type(self).__name__
        try:
            log_evidence = self.log_evidence
        except Exception as e:
            logger.exception(e)
            log_evidence = float("nan")
        return f"{clsname}({self.factor_graph}, " f"log_evidence={log_evidence})"



import pickle

import dill
import pytest

import autofit as af


@pytest.fixture(
    name="optimizer"
)
def make_optimizer():
    return af.DynestyStatic(
        "name"
    )


def test_serialize_optimiser(
        optimizer
):
    optimizer = pickle.loads(
        pickle.dumps(
            optimizer
        )
    )
    assert optimizer.name == "name"


def test_serialize_grid_search(
        optimizer
):
    grid_search = af.SearchGridSearch(
        optimizer
    )
    assert grid_search.logger.name == "GridSearch (name)"
    assert "logger" not in grid_search.__getstate__()

    dumped = dill.dumps(
        grid_search
    )
    loaded = dill.loads(
        dumped
    )
    assert loaded.logger is not None



import numpy as np
import pytest
from scipy import stats

import autofit.mapper.variable
from autofit import graphical as graph


@pytest.fixture(autouse=True)
def set_seed():
    np.random.seed(0)


@pytest.fixture(name="x")
def make_x():
    return autofit.mapper.variable.Variable("x")


@pytest.fixture(name="probit_factor")
def make_probit_factor(x):
    return graph.Factor(stats.norm(loc=0.0, scale=1.0).logcdf, x)



from autofit.database import query as q


class TestString:
    def test_named_query(self):
        query = q.Q(
            "a"
        )

        assert query.query == (
            "SELECT parent_id "
            "FROM object AS o "
            "WHERE o.name = 'a'"
        )

    def test_with_string(self):
        query = q.Q(
            "a",
            q.SV(
                "=",
                'value'
            )
        )

        assert query.query == (
            "SELECT parent_id "
            "FROM object AS o "
            "JOIN string_value AS sv "
            "ON o.id = sv.id "
            "WHERE (o.name = 'a' "
            "AND sv.value = 'value')"
        )

    def test_with_value(self):
        query = q.Q(
            "a",
            q.V(
                "=",
                1
            )
        )

        assert query.query == (
            "SELECT parent_id "
            "FROM object AS o "
            "JOIN value AS v "
            "ON o.id = v.id "
            "WHERE (o.name = 'a' "
            "AND v.value = 1)"
        )

    def test_simple_and(
            self,
            simple_and
    ):
        assert simple_and.query == (
            "SELECT parent_id "
            "FROM object AS o "
            "JOIN value AS v "
            "ON o.id = v.id "
            "WHERE (o.name = 'a' "
            "AND v.value < 1 "
            "AND v.value > 0)"
        )

    def test_simple_or(
            self,
            simple_or
    ):
        assert simple_or.query == (
            "SELECT parent_id "
            "FROM object AS o "
            "JOIN value AS v "
            "ON o.id = v.id "
            "WHERE ((v.value < 1 "
            "OR v.value > 0) "
            "AND o.name = 'a')"
        )

    def test_second_level(
            self,
            second_level
    ):
        assert second_level.query == (
            'SELECT parent_id '
            'FROM object AS o '
            'JOIN value AS v '
            'ON o.id = v.id '
            'WHERE (o.id IN ('
            'SELECT parent_id '
            'FROM object AS o '
            'JOIN value AS v '
            'ON o.id = v.id '
            "WHERE (o.name = 'b' "
            "AND v.value > 0)) "
            "AND o.name = 'a' "
            "AND v.value < 1)"
        )



import pytest

import autofit as af

from autofit.text import formatter as frm

from autofit.mock.mock_model import MockClassx2, MockChildTuplex2


def test_parameterization():
    model = af.Collection(
        collection=af.Collection(
            gaussian=af.Model(af.Gaussian)
        )
    )

    parameterization = model.parameterization
    assert parameterization == (
        """model                                                                                     CollectionPriorModel (N=3)
    collection                                                                            CollectionPriorModel (N=3)
        gaussian                                                                          Gaussian (N=3)"""
    )


def test_root():
    model = af.Model(af.Gaussian)
    parameterization = model.parameterization
    assert parameterization == (
        'model                                                                                     Gaussian (N=3)'
    )


def test_instance():
    model = af.Collection(
        collection=af.Collection(
            gaussian=af.Gaussian()
        )
    )

    parameterization = model.parameterization
    assert parameterization == (
        """model                                                                                     CollectionPriorModel (N=0)
    collection                                                                            CollectionPriorModel (N=0)
        gaussian                                                                          Gaussian (N=0)"""
    )


def test_tuple_prior():
    centre = af.TuplePrior()
    centre.centre_0 = af.UniformPrior()
    centre.centre_1 = af.UniformPrior()

    model = af.Model(
        af.Gaussian,
        centre=centre
    )
    parameterization = model.parameterization
    assert parameterization == (
        'model                                                                                     Gaussian (N=4)'
    )


@pytest.fixture(name="formatter")
def make_info_dict():
    formatter = frm.TextFormatter(line_length=20, indent=4)
    formatter.add(("one", "one"), 1)
    formatter.add(("one", "two"), 2)
    formatter.add(("one", "three", "four"), 4)
    formatter.add(("three", "four"), 4)

    return formatter


class TestGenerateModelInfo:
    def test_info_string(self, formatter):
        ls = formatter.list

        assert ls[0] == "one"
        assert len(ls[1]) == 21
        assert ls[1] == "    one             1"
        assert ls[2] == "    two             2"
        assert ls[3] == "    three"
        assert ls[4] == "        four        4"
        assert ls[5] == "three"
        assert ls[6] == "    four            4"

    def test_basic(self):
        mm = af.ModelMapper()
        mm.mock_class = MockClassx2
        model_info = mm.info

        assert (
                model_info
                == """mock_class
    one                                                                                   UniformPrior, lower_limit = 0.0, upper_limit = 1.0
    two                                                                                   UniformPrior, lower_limit = 0.0, upper_limit = 2.0"""
        )

    def test_with_instance(self):
        mm = af.ModelMapper()
        mm.mock_class = MockClassx2

        mm.mock_class.two = 1.0

        model_info = mm.info

        assert (
                model_info
                == """mock_class
    one                                                                                   UniformPrior, lower_limit = 0.0, upper_limit = 1.0
    two                                                                                   1.0"""
        )

    def test_with_tuple(self):
        mm = af.ModelMapper()
        mm.tuple = (0, 1)

        assert (
                mm.info
                == "tuple                                                                                     (0, 1)"
        )

    # noinspection PyUnresolvedReferences
    def test_tuple_instance_model_info(self, mapper):
        mapper.mock_cls = MockChildTuplex2
        info = mapper.info

        mapper.mock_cls.tup_0 = 1.0

        assert len(mapper.mock_cls.tup.instance_tuples) == 1
        assert len(mapper.mock_cls.instance_tuples) == 1

        assert len(info.split("\n")) == len(mapper.info.split("\n"))



import autofit as af

from autofit.mock.mock_model import MockClassx2

class GaussianChild(af.Gaussian):
    pass


def test_inheritance():
    collection = af.CollectionPriorModel(
        first=af.Model(
            GaussianChild
        ),
        second=GaussianChild()
    )

    assert collection.has_instance(
        af.Gaussian
    )
    assert collection.has_model(
        af.Gaussian
    )


def test_embedded():
    collection = af.Collection(
        model=af.Model(
            af.Gaussian,
            centre=GaussianChild
        )
    )
    assert collection.has_model(
        GaussianChild
    )


def test_no_free_parameters():
    collection = af.CollectionPriorModel(
        gaussian=af.Model(
            af.Gaussian,
            centre=1.0,
            normalization=1.0,
            sigma=1.0,
        )
    )
    assert collection.prior_count == 0
    assert collection.has_model(
        af.Gaussian
    ) is False


def test_instance():
    collection = af.CollectionPriorModel(
        gaussian=af.Gaussian()
    )

    assert collection.has_instance(
        af.Gaussian
    ) is True
    assert collection.has_model(
        af.Gaussian
    ) is False


def test_model():
    collection = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian
        )
    )

    assert collection.has_model(
        af.Gaussian
    ) is True
    assert collection.has_instance(
        af.Gaussian
    ) is False


def test_both():
    collection = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian
        ),
        gaussian_2=af.Gaussian()
    )

    assert collection.has_model(
        af.Gaussian
    ) is True
    assert collection.has_instance(
        af.Gaussian
    ) is True


def test_embedded():
    collection = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian,
            centre=af.Gaussian()
        ),
    )

    assert collection.has_model(
        af.Gaussian
    ) is True
    assert collection.has_instance(
        af.Gaussian
    ) is True


def test_is_only_model():
    collection = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian
        ),
        gaussian_2=af.PriorModel(
            af.Gaussian
        )
    )

    assert collection.is_only_model(
        af.Gaussian
    ) is True

    collection.other = af.PriorModel(
        MockClassx2
    )

    assert collection.is_only_model(
        af.Gaussian
    ) is False



from copy import copy

import pytest

import autofit as af
from autofit import graphical as g
from autofit.graphical.expectation_propagation import FactorHistory, EPHistory


@pytest.fixture(name="success")
def make_success():
    return g.Status(success=True)


@pytest.fixture(name="failure")
def make_failure():
    return g.Status(success=False, updated=False)


def test_truthy_status(success, failure):
    assert success
    assert not failure


def identity(x):
    return x


@pytest.fixture(name="factor")
def make_factor():
    return g.Factor(identity)


@pytest.fixture(name="approx")
def make_approx(factor):
    return g.EPMeanField(
        factor_graph=g.FactorGraph([factor]),
        factor_mean_field={
            factor: g.MeanField(
                {
                    variable: af.GaussianPrior(mean=0, sigma=1)
                    for variable in factor.variables
                }
            )
        },
    )


@pytest.fixture(name="factor_history")
def make_factor_history(factor, approx, success):
    return FactorHistory(factor)


def test_latest(success, factor_history, approx):
    factor_history(approx, success)

    assert factor_history.latest_successful == approx


def test_previous(success, factor_history, approx):
    factor_history(approx, success)
    factor_history(copy(approx), success)

    assert factor_history.previous_successful is approx


def test_failure(success, failure, factor_history, approx):
    approx_2 = copy(approx)

    factor_history(approx, success)
    factor_history(approx_2, success)
    factor_history(copy(approx), failure)

    assert factor_history.latest_successful == approx_2
    assert factor_history.previous_successful == approx


@pytest.fixture(name="trivial_history")
def make_trivial_history(factor_history, approx, success):
    trivial_history = copy(factor_history)
    trivial_history(approx, success)
    trivial_history(approx, success)
    return trivial_history


def test_kl_divergence(trivial_history):
    assert trivial_history.kl_divergence() == 0


def test_evidence_divergence(trivial_history):
    assert trivial_history.evidence_divergence() == 0


def test_default_inf(factor_history):
    assert factor_history.kl_divergence() == float("inf")
    assert factor_history.evidence_divergence() == float("inf")


def test_ep_history(factor, approx, success):
    history = EPHistory(evidence_tol=1e-1)

    history(factor, approx, success)

    assert history.is_kl_converged(factor) is False
    assert history.is_kl_evidence_converged(factor) is False
    assert history.is_kl_converged(factor) is False

    history(factor, approx, success)

    assert history.is_kl_converged(factor)
    assert history.is_kl_evidence_converged(factor)
    assert history.is_kl_converged(factor)


class MockEPMeanField:
    def __init__(self, value):
        self.value = value

    @property
    def log_evidence(self):
        return self.value

    @property
    def mean_field(self):
        return self

    def kl(self, other):
        return self.value - other.value


# noinspection PyTypeChecker
def test_evidences(factor_history, success, failure):
    factor_history(MockEPMeanField(1), success)
    factor_history(MockEPMeanField(2), failure)
    factor_history(MockEPMeanField(3), success)

    assert factor_history.evidences == [1, None, 3]


# noinspection PyTypeChecker
def test_kl_divergences(factor_history, success, failure):
    factor_history(MockEPMeanField(1), success)
    factor_history(MockEPMeanField(2), failure)
    factor_history(MockEPMeanField(3), success)
    factor_history(MockEPMeanField(4), success)

    assert factor_history.kl_divergences == [None, None, 1]



import pytest

import autofit as af
from autofit.mock.mock_model import MockWithTuple
from autofit.mock.mock import MockSearch

@pytest.fixture(
    name="model"
)
def make_model():
    return af.Collection(
        gaussian=af.Model(
            af.Gaussian
        )
    )


class TestInstanceFromPathArguments:
    def test(
            self,
            model
    ):
        instance = model.instance_from_path_arguments({
            ("gaussian", "centre"): 0.1,
            ("gaussian", "normalization"): 0.2,
            ("gaussian", "sigma"): 0.3
        })
        assert instance.gaussian.centre == 0.1
        assert instance.gaussian.normalization == 0.2
        assert instance.gaussian.sigma == 0.3

    def test_prior_linking(
            self,
            model
    ):
        model.gaussian.centre = model.gaussian.normalization
        instance = model.instance_from_path_arguments({
            ("gaussian", "centre",): 0.1,
            ("gaussian", "sigma",): 0.3
        })
        assert instance.gaussian.centre == 0.1
        assert instance.gaussian.normalization == 0.1
        assert instance.gaussian.sigma == 0.3

        instance = model.instance_from_path_arguments({
            ("gaussian", "normalization",): 0.2,
            ("gaussian", "sigma",): 0.3
        })
        assert instance.gaussian.centre == 0.2
        assert instance.gaussian.normalization == 0.2
        assert instance.gaussian.sigma == 0.3


@pytest.fixture(
    name="underscore_model"
)
def make_underscore_model():
    return af.Collection(
        gaussian_component=af.Model(
            af.Gaussian
        )
    )


class TestInstanceFromPriorNames:
    def test(self, model):
        instance = model.instance_from_prior_name_arguments({
            "gaussian_centre": 0.1,
            "gaussian_normalization": 0.2,
            "gaussian_sigma": 0.3
        })
        assert instance.gaussian.centre == 0.1
        assert instance.gaussian.normalization == 0.2
        assert instance.gaussian.sigma == 0.3

    def test_underscored_names(self, underscore_model):
        instance = underscore_model.instance_from_prior_name_arguments({
            "gaussian_component_centre": 0.1,
            "gaussian_component_normalization": 0.2,
            "gaussian_component_sigma": 0.3
        })
        assert instance.gaussian_component.centre == 0.1
        assert instance.gaussian_component.normalization == 0.2
        assert instance.gaussian_component.sigma == 0.3

    def test_prior_linking(self, underscore_model):
        underscore_model.gaussian_component.normalization = (
            underscore_model.gaussian_component.centre
        )
        instance = underscore_model.instance_from_prior_name_arguments({
            "gaussian_component_centre": 0.1,
            "gaussian_component_sigma": 0.3
        })
        assert instance.gaussian_component.centre == 0.1
        assert instance.gaussian_component.normalization == 0.1
        assert instance.gaussian_component.sigma == 0.3

        instance = underscore_model.instance_from_prior_name_arguments({
            "gaussian_component_normalization": 0.2,
            "gaussian_component_sigma": 0.3
        })
        assert instance.gaussian_component.centre == 0.2
        assert instance.gaussian_component.normalization == 0.2
        assert instance.gaussian_component.sigma == 0.3

    def test_path_for_name(self, underscore_model):
        assert underscore_model.path_for_name(
            "gaussian_component_centre"
        ) == (
                   "gaussian_component",
                   "centre"
               )


def test_component_names():
    model = af.Model(
        af.Gaussian
    )
    assert model.model_component_and_parameter_names == [
        'centre', 'normalization', 'sigma'
    ]


def test_with_tuple():
    with_tuple = af.Model(
        MockWithTuple
    )
    assert with_tuple.model_component_and_parameter_names == [
        "tup_0", "tup_1"
    ]


@pytest.fixture(
    name="linked_model"
)
def make_linked_model():
    model = af.Model(
        af.Gaussian
    )
    model.sigma = model.centre
    return model


class TestAllPaths:
    def test_independent(self):
        model = af.Model(
            af.Gaussian
        )

        assert model.all_paths_prior_tuples == [
            ((("centre",),), model.centre),
            ((("normalization",),), model.normalization),
            ((("sigma",),), model.sigma),
        ]

    def test_linked(self, linked_model):
        assert linked_model.all_paths_prior_tuples == [
            ((("centre",), ("sigma",)), linked_model.centre),
            ((("normalization",),), linked_model.normalization)
        ]

    def test_names_independent(self):
        model = af.Model(
            af.Gaussian
        )

        assert model.all_name_prior_tuples == [
            (("centre",), model.centre),
            (("normalization",), model.normalization),
            (("sigma",), model.sigma),
        ]

    def test_names_linked(self, linked_model):
        assert linked_model.all_name_prior_tuples == [
            (("centre", "sigma"), linked_model.centre),
            (("normalization",), linked_model.normalization)
        ]


@pytest.fixture(
    name="samples"
)
def make_samples(model):
    return af.Samples(
        model,
        [
            af.Sample(
                log_likelihood=1.0,
                log_prior=1.0,
                weight=1.0,
                kwargs={
                    ("gaussian", "centre"): 0.1,
                    ("gaussian", "normalization"): 0.2,
                    ("gaussian", "sigma"): 0.3,
                }
            )
        ]
    )


@pytest.mark.parametrize(
    "path, value",
    [
        (("gaussian", "centre"), [0.1]),
        (("gaussian", "normalization"), [0.2]),
        (("gaussian", "sigma"), [0.3]),
    ]
)
def test_values_for_path(
        samples,
        path,
        value
):
    assert samples.values_for_path(
        path
    ) == value


@pytest.fixture(
    name="result"
)
def make_result(
        model,
        samples
):
    return af.Result(
        samples,
        model,
        MockSearch()
    )


@pytest.fixture(
    name="modified_result"
)
def make_modified_result(
        model,
        samples
):
    model.gaussian.sigma = af.GaussianPrior(
        mean=0.5,
        sigma=1
    )
    model.gaussian.centre = af.GaussianPrior(
        mean=0.5,
        sigma=1
    )
    return af.Result(
        samples,
        model,
        MockSearch()
    )


class TestFromResult:
    def test_instance(
            self,
            result
    ):
        instance = result.max_log_likelihood_instance

        assert instance.gaussian.centre == 0.1
        assert instance.gaussian.normalization == 0.2
        assert instance.gaussian.sigma == 0.3

    def test_model(
            self,
            result
    ):
        model = result.model

        assert model.gaussian.centre.mean == 0.1
        assert model.gaussian.normalization.mean == 0.2
        assert model.gaussian.sigma.mean == 0.3

    def test_modified_instance(
            self,
            modified_result
    ):
        instance = modified_result.max_log_likelihood_instance

        assert instance.gaussian.centre == 0.1
        assert instance.gaussian.normalization == 0.2
        assert instance.gaussian.sigma == 0.3

    def test_modified_model(
            self,
            modified_result
    ):
        model = modified_result.model

        assert model.gaussian.centre.mean == 0.1
        assert model.gaussian.normalization.mean == 0.2
        assert model.gaussian.sigma.mean == 0.3



import autofit as af
from autofit import database as db


def test_embedded_query(
        session,
        aggregator
):
    model_1 = db.Fit(
        id="model_1",
        instance=af.Collection(
            gaussian=af.Gaussian(
                centre=1
            )
        ),
        info={"info": 3}
    )
    model_2 = db.Fit(
        id="model_2",
        instance=af.Collection(
            gaussian=af.Gaussian(
                centre=2
            )
        ),
        info={"info": 4}
    )

    session.add_all([
        model_1,
        model_2
    ])

    result = aggregator.query(
        aggregator.model.centre == 0
    )

    assert result == []

    result = aggregator.query(
        aggregator.model.gaussian.centre == 1
    )

    assert result == [model_1]

    result = aggregator.query(
        aggregator.model.gaussian.centre == 2
    )

    assert result == [model_2]


def test_query(
        aggregator,
        gaussian_1,
        gaussian_2
):
    result = aggregator.query(
        aggregator.model.centre == 0
    )

    assert result == []

    result = aggregator.query(
        aggregator.model.centre == 1
    )

    assert result == [gaussian_1]

    result = aggregator.query(
        aggregator.model.centre == 2
    )

    assert result == [gaussian_2]



import autofit as af
from autofit.non_linear.parallel.sneaky import SneakyProcess

from os import path

class MockSneakyProcess(SneakyProcess):
    def __init__(
            self,
            paths,
    ):

        super().__init__(
            name="test",
            paths=paths
        )

def test__test_mode_parallel_profile_outputs_prof_files():

    paths = af.DirectoryPaths(
        path_prefix=path.join("non_linear", "parallel"),
    )

    process = MockSneakyProcess(paths=paths)

    # TODO : I dont know how to make it so run doesn't end up in an infinite loop?

#    process.run()


from .non_linear.grid.grid_search import GridSearch as SearchGridSearch
from . import conf
from . import exc
from .database.aggregator.aggregator import GridSearchAggregator
from .non_linear.samples import MCMCSamples
from .non_linear.samples import NestSamples
from .non_linear.samples import Samples
from .non_linear.samples import PDFSamples
from .non_linear.samples import Sample
from .non_linear.samples import load_from_table
from .non_linear.samples import StoredSamples
from .database.aggregator import Aggregator
from .database.model import Fit
from .database.aggregator import Query
from .database.model.fit import Fit
from .aggregator.search_output import SearchOutput
from .mapper import link
from .mapper import prior
from .mapper.model import AbstractModel
from .mapper.model import ModelInstance
from .mapper.model import ModelInstance as Instance
from .mapper.model import path_instances_of_class
from .mapper.model_mapper import ModelMapper
from .mapper.model_mapper import ModelMapper as Mapper
from .mapper.model_object import ModelObject
from .mapper.prior.assertion import ComparisonAssertion
from .mapper.prior.assertion import ComparisonAssertion
from .mapper.prior.assertion import GreaterThanLessThanAssertion
from .mapper.prior.assertion import GreaterThanLessThanEqualAssertion
from .mapper.prior.deferred import DeferredArgument
from .mapper.prior.deferred import DeferredInstance
from .mapper.prior.width_modifier import AbsoluteWidthModifier
from .mapper.prior.width_modifier import RelativeWidthModifier
from .mapper.prior.width_modifier import WidthModifier
from .mapper.prior.prior import GaussianPrior
from .mapper.prior.prior import LogUniformPrior
from .mapper.prior.abstract import Prior
from .mapper.prior.tuple_prior import TuplePrior
from .mapper.prior.prior import UniformPrior
from .mapper.prior_model.abstract import AbstractPriorModel
from .mapper.prior_model.annotation import AnnotationPriorModel
from .mapper.prior_model.attribute_pair import AttributeNameValue
from .mapper.prior_model.attribute_pair import InstanceNameValue
from .mapper.prior_model.attribute_pair import PriorNameValue
from .mapper.prior_model.attribute_pair import cast_collection
from .mapper.prior_model.collection import CollectionPriorModel
from .mapper.prior_model.collection import CollectionPriorModel as Collection
from .mapper.prior_model.prior_model import PriorModel
from .mapper.prior_model.prior_model import PriorModel as Model
from .mapper.prior_model.util import PriorModelNameValue
from .non_linear.abstract_search import NonLinearSearch
from .non_linear.abstract_search import PriorPasser
from .non_linear.analysis.analysis import Analysis
from .non_linear.grid.grid_search import GridSearchResult
from .non_linear.initializer import InitializerBall
from .non_linear.initializer import InitializerPrior
from .non_linear.mcmc.auto_correlations import AutoCorrelationsSettings
from .non_linear.mcmc.emcee.emcee import Emcee
from .non_linear.mcmc.zeus.zeus import Zeus
from .non_linear.nest.dynesty import DynestyDynamic
from .non_linear.nest.dynesty import DynestyStatic
from .non_linear.nest.ultranest.ultranest import UltraNest
from .non_linear.optimize.drawer.drawer import Drawer
from .non_linear.optimize.lbfgs.lbfgs import LBFGS
from .non_linear.optimize.pyswarms.globe import PySwarmsGlobal
from .non_linear.optimize.pyswarms.local import PySwarmsLocal
from .non_linear.paths import DirectoryPaths
from .non_linear.paths import DatabasePaths
from .non_linear.result import Result
from .non_linear.result import ResultsCollection
from .non_linear.settings import SettingsSearch
from .example.model import Gaussian
from .text import formatter
from .text import samples_text
from .tools import util

from . import example as ex
from . import database as db

conf.instance.register(__file__)

__version__ = '2022.02.14.1'



def test_contain_identifier(
        grid_search,
        grid_paths
):
    for paths in grid_paths:
        identifier = grid_search.paths.identifier
        output_path = paths.output_path
        assert identifier in output_path
        assert not output_path.endswith(
            identifier
        )


def test_does_not_contain_identifier(
        grid_paths
):
    for paths in grid_paths:
        assert paths.identifier not in paths.output_path


def test_distinct_identifiers(
        grid_search,
        grid_paths
):
    identifiers = {
        paths.identifier
        for paths in grid_paths
    }
    assert len(identifiers) == 100
    assert grid_search.paths.identifier not in identifiers



import pytest

import autofit as af
from autofit import database as db


@pytest.fixture(
    name="gaussian_none"
)
def make_gaussian_none(
        session
):
    gaussian_none = db.Fit(
        id="gaussian_none",
        instance=af.Gaussian(
            centre=None
        )
    )
    session.add(gaussian_none)
    session.commit()
    return gaussian_none


def test_query_none(
        gaussian_none,
        aggregator
):
    assert aggregator.query(
        aggregator.model.centre == None
    ) == [gaussian_none]


def test_not(
        gaussian_1,
        gaussian_2,
        aggregator
):
    assert aggregator.query(
        aggregator.model.centre != None
    ) == [gaussian_1, gaussian_2]


def test_combined(
        gaussian_none,
        gaussian_1,
        aggregator
):
    assert aggregator.query(
        (aggregator.model.centre == None) | (aggregator.model.centre == 1.0)
    ) == [gaussian_1, gaussian_none]



from os import path

import pytest

import autofit as af

from autofit.mock.mock import MockSamples
from autofit.mock.mock_model import MockClassx2
from autofit.mock.mock_model import MockClassx2FormatExp
from autofit.non_linear.samples import Sample
from autofit import StoredSamples
from autofit.text import samples_text

text_path = path.join("{}".format(path.dirname(path.realpath(__file__))), "files", "samples")


@pytest.fixture(name="model")
def make_model():
    return af.ModelMapper(mock_class=MockClassx2)


@pytest.fixture(name="samples")
def make_samples(model):
    parameters = [[1.0, 2.0], [1.2, 2.2]]

    log_likelihood_list = [1.0, 0.0]

    return StoredSamples(
        model=model,
        sample_list=Sample.from_lists(
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[0.0, 0.0],
            weight_list=log_likelihood_list,
            model=model
        )
    )


def test__summary(samples):
    results_at_sigma = samples_text.summary(samples=samples, sigma=3.0)

    assert "one       1.00 (1.00, 1.20)" in results_at_sigma
    assert "two       2.00 (2.00, 2.20)" in results_at_sigma


def test__latex(samples):
    latex_results_at_sigma = samples_text.latex(samples=samples, sigma=3.0)

    assert r"one_label^{\rm{o}} = 1.00^{+0.20}_{-0.00} & " in latex_results_at_sigma
    assert r"two_label^{\rm{o}} = 2.00^{+0.20}_{-0.00}" in latex_results_at_sigma

    latex_results_at_sigma = samples_text.latex(samples=samples, sigma=3.0, include_quickmath=True)

    assert r"$one_label^{\rm{o}} = 1.00^{+0.20}_{-0.00}$ & " in latex_results_at_sigma
    assert r"$two_label^{\rm{o}} = 2.00^{+0.20}_{-0.00}$" in latex_results_at_sigma

    model = af.ModelMapper(mock_class=MockClassx2FormatExp)

    parameters = [[1.0, 200.0], [1.2, 200.0]]

    log_likelihood_list = [1.0, 0.0]

    samples_exp = MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            parameter_lists=parameters,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=[0.0, 0.0],
            weight_list=log_likelihood_list,
            model=model
        )
    )

    latex_results_at_sigma = samples_text.latex(samples=samples_exp, sigma=3.0, include_quickmath=True)

    print(latex_results_at_sigma)

    assert r"$one_label^{\rm{o}} = 1.00^{+0.20}_{-0.00}$ & " in latex_results_at_sigma
    assert r"$one_label^{\rm{o}} = 1.00^{+0.20}_{-0.00}$ & $2.00^{+0.00}_{-0.00} \times 10^{2}$" in latex_results_at_sigma


import autofit as af

from autofit.mock.mock_model import MockOverload


def test_constructor():
    prior_model = af.PriorModel(MockOverload)

    assert prior_model.prior_count == 1

    instance = prior_model.instance_from_prior_medians()

    assert instance.one == 1.0
    assert instance.two == 2


def test_alternative():
    prior_model = af.PriorModel(MockOverload.with_two)

    assert prior_model.prior_count == 1

    instance = prior_model.instance_from_prior_medians()

    assert instance.two == 1.0
    assert instance.one == 1.0 / 2



from uuid import uuid4
import numpy as np

import autofit as af


def test_float_inequality(session):
    aggregator = af.Aggregator(session)

    for sigma in [
        0.9992237362814176,
        4.9687212446221904,
        9.967065800134504,
    ]:
        session.add(
            af.db.Fit(
                id=str(uuid4()),
                instance={
                    "gaussian": af.Gaussian(
                        sigma=sigma
                    )
                }
            )
        )
        session.commit()

    assert len(aggregator) == 3

    assert len(aggregator.query(
        aggregator.model.gaussian.sigma < 3
    )) == 1


def test_numpy_values(
        session
):
    aggregator = af.Aggregator(session)
    fit = af.db.Fit(
        id=str(uuid4())
    )
    array = np.zeros(
        (10, 10)
    )
    fit["data"] = array
    session.add(
        fit
    )
    session.commit()

    assert (
            aggregator.values(
                "data"
            )[0] == array
    ).all()



import pytest

import autofit as af
from autofit import database as m
from autofit.mock.mock import MockSamples


@pytest.fixture(
    name="save_samples",
    autouse=True
)
def save_samples(
        paths,
        sample
):
    samples =MockSamples(
        model=af.Model(
            af.Gaussian
        ),

    )
    samples.sample_list = [sample]
    paths.save_samples(
        samples
    )


@pytest.fixture(
    name="sample"
)
def make_sample():
    return af.Sample(
        log_likelihood=1.0,
        log_prior=1.0,
        weight=0.5,
        kwargs=dict(
            centre=1.0,
            normalization=2.0,
            sigma=3.0
        )
    )


def test_serialise_sample(sample):
    assert isinstance(
        sample.kwargs,
        dict
    )

    sample = m.Object.from_object(
        sample
    )()
    assert "centre" in sample.kwargs


def test_load_samples(
        paths
):
    samples = paths._load_samples()

    assert samples.model.cls is af.Gaussian

    sample, = samples.sample_list
    assert sample.weight == 0.5



import pickle

from autofit.messages.transform import log_10_transform


def test_pickle_log_10_transform():
    loaded = pickle.loads(
        pickle.dumps(
            log_10_transform
        )
    )
    assert loaded.transform(10) == 1



import pytest
from scipy import stats

from autofit import graphical as graph
from autofit.messages.normal import NormalMessage

x = graph.Variable("x")
graph.Factor(stats.norm(loc=-0.5, scale=0.5).logpdf, x)


@pytest.fixture(name="normal_factor")
def make_normal_factor(x):
    return graph.Factor(stats.norm(loc=-0.5, scale=0.5).logpdf, x)


@pytest.fixture(name="model")
def make_model(probit_factor, normal_factor):
    return probit_factor * normal_factor


@pytest.fixture(name="message")
def make_message():
    return NormalMessage(0, 1)


@pytest.fixture(name="model_approx")
def make_model_approx(model, x, message):
    return graph.EPMeanField.from_kws(model, {x: message})


@pytest.fixture(name="probit_approx")
def make_probit_approx(probit_factor, model_approx):
    return model_approx.factor_approximation(probit_factor)


def test_approximations(probit_approx, model_approx, x, message):
    opt = graph.LaplaceOptimiser()
    probit_model_dist, status = opt.optimise_approx(probit_approx)

    # get updated factor approximation
    probit_project, status = probit_approx.project(probit_model_dist, delta=1.0)

    assert probit_project.model_dist[x].mean == pytest.approx(0.506, rel=0.1)
    assert probit_project.model_dist[x].sigma == pytest.approx(0.814, rel=0.1)

    assert probit_project.factor_dist[x].mean == pytest.approx(1.499, rel=0.1)
    assert probit_project.factor_dist[x].sigma == pytest.approx(1.401, rel=0.1)



def test_completed_aggregator(
        aggregator
):
    aggregator = aggregator(
        aggregator.search.is_complete
    )
    assert len(aggregator) == 1


class TestLoading:
    def test_unzip(self, aggregator):
        assert len(aggregator) == 2

    def test_pickles(self, aggregator):
        assert list(aggregator.values("dataset"))[0]["name"] == "dataset"


class TestOperations:
    def test_attribute(self, aggregator):
        assert list(aggregator.values("pipeline")) == [
            "pipeline0",
            "pipeline1"
        ]

    def test_indexing(self, aggregator):
        assert list(aggregator[1:].values("pipeline")) == ["pipeline1"]
        assert list(aggregator[:1].values("pipeline")) == ["pipeline0"]
        assert list(aggregator[1: 2].values("pipeline")) == ["pipeline1"]
        assert list(aggregator[0: 1].values("pipeline")) == ["pipeline0"]
        assert list(aggregator[-1:].values("pipeline")) == ["pipeline1"]
        assert list(aggregator[:-1].values("pipeline")) == ["pipeline0"]
        assert aggregator[0]["pipeline"] == "pipeline0"
        assert aggregator[-1]["pipeline"] == "pipeline1"

    def test_map(self, aggregator):
        def some_function(fit):
            return f"{fit.id} {fit['pipeline']}"

        results = aggregator.map(some_function)
        assert list(results) == [
            'complete pipeline0',
            'incomplete pipeline1'
        ]



import numpy as np
import pytest

import autofit as af
import autofit.graphical as ep
from test_autofit.graphical.gaussian.model import Gaussian, make_data, Analysis


@pytest.fixture(name="make_model_factor")
def make_make_model_factor(normalization, normalization_prior, x):
    def make_factor_model(
            centre: float, sigma: float, optimiser=None
    ) -> ep.AnalysisFactor:
        """
        We'll make a LikelihoodModel for each Gaussian we're fitting.

        First we'll make the actual data to be fit.

        Note that the normalization value is shared.
        """
        y = make_data(
            Gaussian(centre=centre, normalization=normalization, sigma=sigma), x
        )

        """
        Next we need a prior model.
    
        Note that the normalization prior is shared.
        """
        prior_model = af.PriorModel(
            Gaussian,
            centre=af.GaussianPrior(mean=50, sigma=20),
            normalization=normalization_prior,
            sigma=af.GaussianPrior(mean=10, sigma=10),
        )

        """
        Finally we combine the likelihood function with the prior model to produce a likelihood
        factor - this will be converted into a ModelFactor which is like any other factor in the
        factor graph.
        
        We can also pass a custom optimiser in here that will be used to fit the factor instead
        of the default optimiser.
        """
        return ep.AnalysisFactor(
            prior_model, analysis=Analysis(x=x, y=y), optimiser=optimiser
        )

    return make_factor_model


@pytest.fixture(name="normalization")
def make_normalization():
    return 25.0


@pytest.fixture(name="normalization_prior")
def make_normalization_prior():
    return af.GaussianPrior(mean=25, sigma=10)


@pytest.fixture(name="factor_model")
def make_factor_model_collection(make_model_factor):
    """
    Here's a good example in which we have two Gaussians fit with a shared variable

    We have a shared normalization value and a shared normalization prior

    Multiplying together multiple LikelihoodModels gives us a factor model.

    The factor model can compute all the variables and messages required as well as construct
    a factor graph representing a fit on the ensemble.
    """
    return ep.FactorGraphModel(
        make_model_factor(centre=40, sigma=10), make_model_factor(centre=60, sigma=15)
    )


def test_custom_optimiser(make_model_factor):
    other_optimiser = ep.LaplaceOptimiser()

    factor_1 = make_model_factor(centre=40, sigma=10, optimiser=other_optimiser)
    factor_2 = make_model_factor(centre=60, sigma=15)

    factor_model = ep.FactorGraphModel(factor_1, factor_2)

    default_optimiser = ep.LaplaceOptimiser()
    ep_optimiser = factor_model._make_ep_optimiser(default_optimiser)

    factor_optimisers = ep_optimiser.factor_optimisers
    assert factor_optimisers[factor_1] is other_optimiser
    assert factor_optimisers[factor_2] is default_optimiser


def test_factor_model_attributes(factor_model):
    """
    There are:
    - 5 messages - one for each prior
    - 7 factors - one for each prior plus one for each likelihood
    """
    assert len(factor_model.message_dict) == 5
    assert len(factor_model.graph.factors) == 7


def _test_optimise_factor_model(factor_model):
    """
    We optimise the model
    """
    laplace = ep.LaplaceOptimiser()

    collection = factor_model.optimise(laplace)

    """
    And what we get back is actually a PriorModelCollection
    """
    assert 25.0 == pytest.approx(collection[0].normalization.mean, rel=0.1)
    assert collection[0].normalization is collection[1].normalization


def test_gaussian():
    n_observations = 100
    x = np.arange(n_observations)
    y = make_data(Gaussian(centre=50.0, normalization=25.0, sigma=10.0), x)

    prior_model = af.PriorModel(
        Gaussian,
        centre=af.GaussianPrior(mean=50, sigma=20),
        normalization=af.GaussianPrior(mean=25, sigma=10),
        sigma=af.GaussianPrior(mean=10, sigma=10),
    )

    factor_model = ep.AnalysisFactor(prior_model, analysis=Analysis(x=x, y=y))

    laplace = ep.LaplaceOptimiser()
    model = factor_model.optimise(laplace)

    assert model.centre.mean == pytest.approx(50, rel=0.1)
    assert model.normalization.mean == pytest.approx(25, rel=0.1)
    assert model.sigma.mean == pytest.approx(10, rel=0.1)


@pytest.fixture(name="prior_model")
def make_prior_model():
    return af.PriorModel(Gaussian)


@pytest.fixture(name="likelihood_model")
def make_factor_model(prior_model):
    class MockAnalysis(af.Analysis):
        @staticmethod
        def log_likelihood_function(*_):
            return 1

    return ep.AnalysisFactor(prior_model, analysis=MockAnalysis())


def test_messages(likelihood_model):
    assert len(likelihood_model.message_dict) == 3


def test_graph(likelihood_model):
    graph = likelihood_model.graph
    assert len(graph.factors) == 4


def test_prior_model_node(likelihood_model):
    prior_model_node = likelihood_model.graph

    result = prior_model_node(
        {variable: np.array([0.5]) for variable in prior_model_node.variables}
    )

    assert isinstance(result, ep.FactorValue)



import pytest

from autofit.database.migration import Step, Migrator


@pytest.fixture(
    name="step_1"
)
def make_step_1():
    return Step(
        "INSERT INTO test (id) VALUES (1)"
    )


@pytest.fixture(
    name="step_2"
)
def make_step_2():
    return Step(
        "INSERT INTO test (id) VALUES (2)"
    )


@pytest.fixture(
    name="migrator"
)
def make_migrator(
        step_1,
        step_2
):
    return Migrator(
        step_1,
        step_2
    )


@pytest.fixture(
    name="revision_1"
)
def make_revision_1(migrator):
    return list(
        migrator.revisions
    )[0]


@pytest.fixture(
    name="revision_2"
)
def make_revision_2(migrator):
    return list(
        migrator.revisions
    )[1]



import autofit as af


def test_save_instance(
        session
):
    paths = af.DatabasePaths(
        session=session
    )
    paths.save_named_instance(
        "name",
        af.Gaussian()
    )
    assert isinstance(
        paths.fit.named_instances[
            "name"
        ],
        af.Gaussian

    )


def test_paths_type(
        database_paths
):
    for path in database_paths:
        assert isinstance(
            path,
            af.DatabasePaths
        )


def test_name_prefix_tag(
        session
):
    paths = af.DatabasePaths(
        session,
        name="name",
        path_prefix="prefix",
        unique_tag="tag"
    )

    fit = paths.fit
    assert fit.name == "name"
    assert fit.path_prefix == "prefix"
    assert fit.unique_tag == "tag"



import numpy as np
import pytest

import autofit as af
import autofit.graphical as ep
from test_autofit.graphical.gaussian.model import Gaussian, make_data, Analysis


@pytest.fixture(name="make_model_factor")
def make_make_model_factor(normalization, normalization_prior, x):
    def make_factor_model(
            centre: float, sigma: float, optimiser=None
    ) -> ep.AnalysisFactor:
        y = make_data(
            Gaussian(centre=centre, normalization=normalization, sigma=sigma), x
        )

        prior_model = af.PriorModel(
            Gaussian,
            centre=af.UniformPrior(lower_limit=10, upper_limit=100),
            normalization=normalization_prior,
            sigma=af.UniformPrior(lower_limit=0, upper_limit=20),
        )

        return ep.AnalysisFactor(
            prior_model, analysis=Analysis(x=x, y=y), optimiser=optimiser
        )

    return make_factor_model


@pytest.fixture(name="normalization")
def make_normalization():
    return 25.0


@pytest.fixture(name="normalization_prior")
def make_normalization_prior():
    return af.UniformPrior(lower_limit=15, upper_limit=35)


@pytest.fixture(name="factor_model")
def make_factor_model_collection(make_model_factor):
    return ep.FactorGraphModel(
        make_model_factor(centre=40, sigma=10), make_model_factor(centre=60, sigma=15)
    )


def test_uniform_edge():
    uniform_prior = af.UniformPrior(lower_limit=10, upper_limit=20)
    assert not np.isnan(uniform_prior.logpdf(10))
    assert not np.isnan(uniform_prior.logpdf(20))


def _test_optimise_factor_model(factor_model):
    laplace = ep.LaplaceOptimiser()

    collection = factor_model.optimise(laplace)

    assert 25.0 == pytest.approx(collection[0].normalization.mean, rel=0.1)
    assert collection[0].normalization is collection[1].normalization


def test_trivial():
    prior = af.UniformPrior(lower_limit=10, upper_limit=20)

    prior_model = af.Collection(value=prior)

    class TrivialAnalysis(af.Analysis):
        def log_likelihood_function(self, instance):
            result = -((instance.value - 14) ** 2)
            return result

    factor_model = ep.AnalysisFactor(prior_model, analysis=TrivialAnalysis())

    optimiser = ep.LaplaceOptimiser()
    # optimiser = af.DynestyStatic()
    model = factor_model.optimise(optimiser)

    assert model.value.mean == pytest.approx(14, rel=0.1)


def _test_gaussian():
    n_observations = 100
    x = np.arange(n_observations)
    y = make_data(Gaussian(centre=50.0, normalization=25.0, sigma=10.0), x)

    prior_model = af.PriorModel(
        Gaussian,
        # centre=af.GaussianPrior(mean=50, sigma=10),
        # normalization=af.GaussianPrior(mean=25, sigma=10),
        sigma=af.GaussianPrior(mean=10, sigma=10),
        centre=af.UniformPrior(lower_limit=30, upper_limit=70),
        normalization=af.UniformPrior(lower_limit=15, upper_limit=35),
        # sigma=af.UniformPrior(lower_limit=5, upper_limit=15),
    )

    factor_model = ep.AnalysisFactor(prior_model, analysis=Analysis(x=x, y=y))

    # optimiser = ep.LaplaceOptimiser(
    #     transform_cls=DiagonalMatrix
    # )
    optimiser = af.DynestyStatic()
    model = factor_model.optimise(optimiser)

    assert model.centre.mean == pytest.approx(50, rel=0.1)
    assert model.normalization.mean == pytest.approx(25, rel=0.1)
    assert model.sigma.mean == pytest.approx(10, rel=0.1)



import logging
import os
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import Dict, Tuple, Optional, List

import matplotlib.pyplot as plt

from autofit import exc
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.expectation_propagation.history import EPHistory
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.graphical.mean_field import MeanField, FactorApproximation, Status
from autofit.graphical.utils import StatusFlag, LogWarnings
from autofit.mapper.identifier import Identifier
from autofit.non_linear.paths import DirectoryPaths
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.tools.util import IntervalCounter

logger = logging.getLogger(__name__)


class AbstractFactorOptimiser(ABC):
    """
    An optimiser used to optimise individual factors during EPOptimisation.
    """

    def __init__(self, initial_values=None, deltas=None):
        self.initial_values = initial_values or {}

        self.deltas = defaultdict(lambda: 1)
        if deltas:
            self.deltas.update(deltas)

    def update_model_approx(
            self,
            new_model_dist: MeanField,
            factor_approx: FactorApproximation,
            model_approx: EPMeanField,
            status: Optional[Status] = Status(),
    ) -> Tuple[EPMeanField, Status]:
        delta = self.deltas[factor_approx.factor]
        projection, status = factor_approx.project(
            new_model_dist, delta=delta, status=status
        )
        new_approx, status = model_approx.project(projection, status)
        return new_approx, status

    @abstractmethod
    def optimise(
            self, factor: Factor, model_approx: EPMeanField, status: Status = Status()
    ) -> Tuple[EPMeanField, Status]:
        pass


class Visualise:
    def __init__(self, ep_history: EPHistory, output_path: Path):
        """
        Handles visualisation of expectation propagation optimisation.

        This includes plotting key metrics such as Evidence and KL Divergence
        which are expected to converge.

        Parameters
        ----------
        ep_history
            A history describing previous optimisations by factor
        output_path
            The path that plots are written to
        """
        self.ep_history = ep_history
        self.output_path = output_path

    def __call__(self):
        """
        Save a plot of Evidence and KL Divergence for the ep_history
        """
        fig, (evidence_plot, kl_plot) = plt.subplots(2)
        fig.suptitle("Evidence and KL Divergence")
        evidence_plot.plot(self.ep_history.evidences(), label="evidence")
        kl_plot.semilogy(self.ep_history.kl_divergences(), label="KL divergence")
        # for factor, factor_history in self.ep_history.items():
        #     evidence_plot.plot(
        #         factor_history.evidences, label=f"{factor.name} evidence"
        #     )
        #     kl_plot.plot(
        #         factor_history.kl_divergences, label=f"{factor.name} divergence"
        #     )
        evidence_plot.legend()
        kl_plot.legend()
        plt.savefig(str(self.output_path / "graph.png"))


class EPOptimiser:
    def __init__(
            self,
            factor_graph: FactorGraph,
            default_optimiser: Optional[AbstractFactorOptimiser] = None,
            factor_optimisers: Optional[Dict[Factor, AbstractFactorOptimiser]] = None,
            ep_history: Optional[EPHistory] = None,
            factor_order: Optional[List[Factor]] = None,
            paths: AbstractPaths = None,
    ):
        """
        Optimise a factor graph.

        Cycles through factors optimising them individually using priors
        created through expectation propagation; in effect the prior for each
        variable for a given factor is the product of the posteriors of that
        variable for all other factors.

        Parameters
        ----------
        factor_graph
            A graph describing the relationships between multiple factors
        default_optimiser
            An optimiser that is used if no specific optimiser is provided for a factor
        factor_optimisers
            Specific optimisers used to optimise each factor
        ep_history
            Optionally specify an alternate history
        factor_order
            The factors in the graph but placed in the order in which they should
            be optimised
        paths
            Optionally define how data should be output
        """
        self.paths = paths or DirectoryPaths(identifier=str(Identifier(factor_graph)))

        factor_optimisers = factor_optimisers or {}
        self.factor_graph = factor_graph
        self.factors = factor_order or self.factor_graph.factors
        self.default_optimiser = default_optimiser

        if default_optimiser is None:
            self.factor_optimisers = factor_optimisers
            missing = set(self.factors) - self.factor_optimisers.keys()
            if missing:
                raise (
                    ValueError(
                        f"missing optimisers for {missing}, "
                        "pass a default_optimiser or add missing optimsers"
                    )
                )
        else:
            self.factor_optimisers = {
                factor: factor_optimisers.get(factor, default_optimiser)
                for factor in self.factors
            }

        for optimiser in self.factor_optimisers.values():
            optimiser.paths = self.paths

        self.ep_history = ep_history or EPHistory()

        with open(self.output_path / "graph.info", "w+") as f:
            f.write(self.factor_graph.info)

        self.visualiser = Visualise(self.ep_history, self.output_path)

    @property
    def output_path(self) -> Path:
        """
        The path at which data will be output. Uses the name of the optimiser.

        If the path does not exist it is created.
        """
        path = Path(self.paths.output_path)
        os.makedirs(path, exist_ok=True)
        return path

    def _log_factor(self, factor: Factor):
        """
        Log information for the factor and its history.
        """
        factor_logger = logging.getLogger(factor.name)
        try:
            factor_history = self.ep_history[factor]
            if factor_history.history:
                log_evidence = factor_history.latest_update.log_evidence
                divergence = factor_history.kl_divergence()

                factor_logger.info(f"Log Evidence = {log_evidence}")
                factor_logger.info(f"KL Divergence = {divergence}")
        except exc.HistoryException as e:
            factor_logger.exception(e)

    def run(
            self,
            model_approx: EPMeanField,
            max_steps: int = 100,
            log_interval: int = 10,
            visualise_interval: int = 100,
            output_interval: int = 10,
    ) -> EPMeanField:
        """
        Run the optimisation on an approximation of the model.

        Parameters
        ----------
        model_approx
            A collection of messages describing priors on the model's variables.
        max_steps
            The maximum number of steps prior to termination. Termination may also
            occur when difference in log evidence or KL Divergence drop below a given
            threshold for two consecutive optimisations of a given factor.
        log_interval
            How steps should we wait before logging information?
        visualise_interval
            How steps should we wait before visualising information?
            This includes plots of KL Divergence and Evidence.
        output_interval
            How steps should we wait before outputting information?
            This includes the model.results file which describes the current mean values
            of each message.

        Returns
        -------
        An updated approximation of the model
        """
        should_log = IntervalCounter(log_interval)
        should_visualise = IntervalCounter(visualise_interval)
        should_output = IntervalCounter(output_interval)

        for _ in range(max_steps):
            for factor, optimiser in self.factor_optimisers.items():
                factor_logger = logging.getLogger(factor.name)
                factor_logger.debug("Optimising...")
                try:
                    with LogWarnings(logger=factor_logger.debug, action='always') as caught_warnings:
                        model_approx, status = optimiser.optimise(
                            factor,
                            model_approx,
                        )

                    messages = status.messages
                    for m in caught_warnings.messages:
                        messages += f"optimise_quasi_newton warning: {m}",

                    status = Status(status.success, messages, status.flag)

                except (ValueError, ArithmeticError, RuntimeError) as e:
                    logger.exception(e)
                    status = Status(
                        False,
                        (f"Factor: {factor} experienced error {e}",),
                        StatusFlag.FAILURE,
                    )

                if status and should_log():
                    self._log_factor(factor)

                factor_logger.debug(status)

                if self.ep_history(factor, model_approx, status):
                    logger.info("Terminating optimisation")
                    break  # callback controls convergence

            else:  # If no break do next iteration
                if should_visualise():
                    self.visualiser()
                if should_output():
                    self._output_results(model_approx)
                continue
            break  # stop iterations

        self.visualiser()
        self._output_results(model_approx)

        return model_approx

    def _output_results(self, model_approx: EPMeanField):
        """
        Save the graph.results text
        """
        with open(self.output_path / "graph.results", "w+") as f:
            f.write(self.factor_graph.make_results_text(model_approx))



import autofit as af
from autofit import database as db


def test_instance_from_prior_medians():
    model = af.Model(
        af.Gaussian
    )
    db.Object.from_object(
        model
    )()
    db.Object.from_object(
        af.Gaussian()
    )()
    instance = model.instance_from_prior_medians()
    db.Object.from_object(
        instance
    )()



import multiprocessing
import os
import shutil
import sys
from os import path
from pathlib import Path

import pytest
from matplotlib import pyplot
from autofit.database.model import sa

from autoconf import conf
from autofit import database as db
from autofit.mock import fixtures

if sys.platform == 'darwin':
    multiprocessing.set_start_method('forkserver')

directory = Path(__file__).parent


@pytest.fixture(
    name="test_directory",
    scope="session"
)
def make_test_directory():
    return directory


@pytest.fixture(
    name="output_directory",
    scope="session"
)
def make_output_directory(
        test_directory
):
    return test_directory / "output"


@pytest.fixture(
    autouse=True,
    scope="session"
)
def remove_output(
        output_directory
):
    yield
    for item in os.listdir(output_directory):
        if item != "non_linear":
            item_path = output_directory / item
            if item_path.is_dir():
                shutil.rmtree(
                    item_path
                )
            else:
                os.remove(
                    item_path
                )


class PlotPatch:
    def __init__(self):
        self.paths = []

    def __call__(self, path, *args, **kwargs):
        self.paths.append(path)


@pytest.fixture(name="plot_patch")
def make_plot_patch(monkeypatch):
    plot_patch = PlotPatch()
    monkeypatch.setattr(pyplot, "savefig", plot_patch)
    return plot_patch


@pytest.fixture(name="session")
def make_session():
    engine = sa.create_engine('sqlite://')
    session = sa.orm.sessionmaker(bind=engine)()
    db.Base.metadata.create_all(engine)
    yield session
    session.close()
    engine.dispose()


@pytest.fixture(
    autouse=True,
    scope="session"
)
def remove_logs():
    yield
    for d, _, files in os.walk(directory):
        for file in files:
            if file.endswith(".log"):
                os.remove(path.join(d, file))


@pytest.fixture(autouse=True)
def set_config_path():
    conf.instance.push(
        new_path=path.join(directory, "config"),
        output_path=path.join(directory, "output")
    )


@pytest.fixture(
    name="model_gaussian_x1"
)
def make_model_gaussian_x1():
    return fixtures.make_model_gaussian_x1()

@pytest.fixture(name="samples_x5")
def make_samples_x5():
    return fixtures.make_samples_x5()




import pytest

import autofit as af


@pytest.fixture(
    name="frozen_model"
)
def make_frozen_model():
    model = af.PriorModel(
        af.Gaussian
    )

    model.freeze()
    return model


@pytest.fixture(
    name="frozen_collection"
)
def make_frozen_collection():
    model = af.CollectionPriorModel(
        gaussian=af.PriorModel(
            af.Gaussian
        )
    )

    model.freeze()
    return model


def test_instantiate(
        frozen_model
):
    frozen_model.instance_from_prior_medians()


def test_modify(
        frozen_model
):
    with pytest.raises(
            AssertionError
    ):
        frozen_model.key = "value"


class TestFrozenCollection:
    def test_append(
            self,
            frozen_collection
    ):
        with pytest.raises(
                AssertionError
        ):
            frozen_collection.append(
                af.Gaussian()
            )

    def test_set_item(
            self,
            frozen_collection
    ):
        with pytest.raises(
                AssertionError
        ):
            frozen_collection["key"] = "value"

    def test_children_frozen(
            self,
            frozen_collection
    ):
        with pytest.raises(
                AssertionError
        ):
            frozen_collection[0].key = "value"


def test_unfreeze(
        frozen_collection
):
    frozen_collection.unfreeze()
    frozen_collection["key"] = "value"
    frozen_collection[0].key = "value"



from abc import ABC, abstractmethod
from functools import reduce
from inspect import getfullargspec
from numbers import Real
from operator import and_
from typing import Dict, Tuple, Iterator
from typing import Optional, Union, Type, List

import numpy as np

from autoconf import cached_property
from autofit.mapper.prior.abstract import Prior
from .transform import AbstractDensityTransform, LinearShiftTransform
from ..mapper.variable import Variable

enforce_id_match = True


class AbstractMessage(Prior, ABC):
    log_base_measure: float
    _projection_class: Optional[Type["AbstractMessage"]] = None
    _multivariate: bool = False
    _parameter_support: Optional[Tuple[Tuple[float, float], ...]] = None
    _support: Optional[Tuple[Tuple[float, float], ...]] = None

    def __init__(self, *parameters: Union[np.ndarray, float], log_norm=0.0, **kwargs):
        super().__init__(**kwargs)
        self.log_norm = log_norm
        self._broadcast = np.broadcast(*parameters)

        if self.shape:
            self.parameters = tuple(np.asanyarray(p) for p in parameters)
        else:
            self.parameters = tuple(parameters)

    def __bool__(self):
        return True

    @cached_property
    @abstractmethod
    def natural_parameters(self):
        pass

    @abstractmethod
    def sample(self, n_samples: Optional[int] = None):
        pass

    @staticmethod
    @abstractmethod
    def invert_natural_parameters(
            natural_parameters: np.ndarray,
    ) -> Tuple[np.ndarray, ...]:
        pass

    @staticmethod
    @abstractmethod
    def to_canonical_form(x: np.ndarray) -> np.ndarray:
        pass

    @cached_property
    @abstractmethod
    def log_partition(self) -> np.ndarray:
        pass

    @cached_property
    @abstractmethod
    def variance(self) -> np.ndarray:
        pass

    @cached_property
    def scale(self) -> np.ndarray:
        return self.scale

    @cached_property
    def std(self) -> np.ndarray:
        return self.variance ** 0.5

    def __hash__(self):
        return self.id

    @classmethod
    def calc_log_base_measure(cls, x):
        return cls.log_base_measure

    def __iter__(self) -> Iterator[np.ndarray]:
        return iter(self.parameters)

    @property
    def shape(self) -> Tuple[int, ...]:
        return self._broadcast.shape

    @property
    def size(self) -> int:
        return self._broadcast.size

    @property
    def ndim(self) -> int:
        return self._broadcast.ndim

    @classmethod
    def from_natural_parameters(
            cls, natural_parameters: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        cls_ = cls._projection_class or cls
        args = cls_.invert_natural_parameters(natural_parameters)
        return cls_(*args, **kwargs)

    @classmethod
    @abstractmethod
    def invert_sufficient_statistics(
            cls, sufficient_statistics: np.ndarray
    ) -> np.ndarray:
        pass

    @classmethod
    def from_sufficient_statistics(
            cls, suff_stats: np.ndarray, **kwargs
    ) -> "AbstractMessage":
        natural_params = cls.invert_sufficient_statistics(suff_stats)
        cls_ = cls._projection_class or cls
        return cls_.from_natural_parameters(natural_params, **kwargs)

    def sum_natural_parameters(self, *dists: "AbstractMessage") -> "AbstractMessage":
        """return the unnormalised result of multiplying the pdf
        of this distribution with another distribution of the same
        type
        """
        new_params = sum(
            (
                dist.natural_parameters
                for dist in self._iter_dists(dists)
                if isinstance(dist, AbstractMessage)
            ),
            self.natural_parameters,
        )
        mul_dist = self.from_natural_parameters(new_params, id_=self.id)
        return mul_dist

    def sub_natural_parameters(self, other: "AbstractMessage") -> "AbstractMessage":
        """return the unnormalised result of dividing the pdf
        of this distribution with another distribution of the same
        type"""
        log_norm = self.log_norm - other.log_norm
        new_params = self.natural_parameters - other.natural_parameters
        div_dist = self.from_natural_parameters(
            new_params, log_norm=log_norm, id_=self.id
        )
        return div_dist

    _multiply = sum_natural_parameters
    _divide = sub_natural_parameters

    def __mul__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, Prior):
            return self._multiply(other)
        else:
            log_norm = self.log_norm + np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm, id_=self.id)

    def __rmul__(self, other: "AbstractMessage") -> "AbstractMessage":
        return self * other

    def __truediv__(self, other: Union["AbstractMessage", Real]) -> "AbstractMessage":
        if isinstance(other, Prior):
            return self._divide(other)
        else:
            log_norm = self.log_norm - np.log(other)
            return type(self)(*self.parameters, log_norm=log_norm, id_=self.id)

    def __pow__(self, other: Real) -> "AbstractMessage":
        natural = self.natural_parameters
        new_params = other * natural
        log_norm = other * self.log_norm
        return self.from_natural_parameters(new_params, log_norm=log_norm, id_=self.id)

    @classmethod
    def parameter_names(cls):
        return getfullargspec(cls.__init__).args[1:-1]

    def __str__(self) -> str:
        param_attrs = [
            (attr, np.asanyarray(getattr(self, attr)))
            for attr in self.parameter_names()
        ]
        if self.shape:
            pad = max(len(attr) for attr, _ in param_attrs)
            attr_str = "    {:<%d}={}" % pad
            param_strs = ",\n".join(
                attr_str.format(attr, np.array2string(val, prefix=" " * (pad + 5)))
                for attr, val in param_attrs
            )
            return f"{type(self).__name__}(\n{param_strs})"
        else:
            param_strs = ", ".join(
                attr + "=" + np.array2string(val, prefix=" " * (len(attr) + 1))
                for attr, val in param_attrs
            )
            return f"{type(self).__name__}({param_strs})"

    __repr__ = __str__

    def pdf(self, x: np.ndarray) -> np.ndarray:
        return np.exp(self.logpdf(x))

    def _broadcast_natural_parameters(self, x):
        shape = np.shape(x)
        if shape == self.shape:
            return self.natural_parameters
        elif shape[1:] == self.shape:
            return self.natural_parameters[:, None, ...]
        else:
            raise ValueError(
                f"shape of passed value {shape} does not "
                f"match message shape {self.shape}"
            )

    def factor(self, x):
        self.assert_within_limits(x)
        return self.logpdf(x)

    def logpdf(self, x: np.ndarray) -> np.ndarray:
        eta = self._broadcast_natural_parameters(x)
        t = self.to_canonical_form(x)
        log_base = self.calc_log_base_measure(x)
        return self.natural_logpdf(eta, t, log_base, self.log_partition)

    @classmethod
    def natural_logpdf(cls, eta, t, log_base, log_partition):
        eta_t = np.multiply(eta, t).sum(0)
        return np.nan_to_num(log_base + eta_t - log_partition, nan=-np.inf)

    def numerical_logpdf_gradient(
            self, x: np.ndarray, eps: float = 1e-6
    ) -> Tuple[np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            logl0 = self.logpdf(x0)
            if self._multivariate:
                grad_logl = np.empty(logl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(logl0.ndim))
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0)
                        grad_logl[sl + it.multi_index] = (logl - logl0) / eps
                        xv -= eps
            else:
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        logl = self.logpdf(x0).sum()  # type: ignore
                        grad_logl[it.multi_index] = (logl - l0) / eps
                        xv -= eps
        else:
            logl0 = self.logpdf(x)
            grad_logl = (self.logpdf(x + eps) - logl0) / eps

        return logl0, grad_logl

    def numerical_logpdf_gradient_hessian(
            self, x: np.ndarray, eps: float = 1e-6
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        shape = np.shape(x)
        if shape:
            x0 = np.array(x, dtype=np.float64)
            if self._multivariate:
                logl0, gradl0 = self.numerical_logpdf_gradient(x0)
                hess_logl = np.empty(gradl0.shape + x0.shape)
                sl = tuple(slice(None) for _ in range(gradl0.ndim))
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        _, gradl = self.numerical_logpdf_gradient(x0)
                        hess_logl[sl + it.multi_index] = (gradl - gradl0) / eps
                        xv -= eps
            else:
                logl0 = self.logpdf(x0)
                l0 = logl0.sum()
                grad_logl = np.empty_like(x0)
                hess_logl = np.empty_like(x0)
                with np.nditer(x0, flags=["multi_index"], op_flags=["readwrite"]) as it:
                    for xv in it:
                        xv += eps
                        l1 = self.logpdf(x0).sum()
                        xv -= 2 * eps
                        l2 = self.logpdf(x0).sum()
                        g1 = (l1 - l0) / eps
                        g2 = (l0 - l2) / eps
                        grad_logl[it.multi_index] = g1
                        hess_logl[it.multi_index] = (g1 - g2) / eps
                        xv += eps

                gradl0 = grad_logl
        else:
            logl0 = self.logpdf(x)
            logl1 = self.logpdf(x + eps)
            logl2 = self.logpdf(x - eps)
            gradl0 = (logl1 - logl0) / eps
            gradl1 = (logl0 - logl2) / eps
            hess_logl = (gradl0 - gradl1) / eps

        return logl0, gradl0, hess_logl

    logpdf_gradient = numerical_logpdf_gradient
    logpdf_gradient_hessian = numerical_logpdf_gradient_hessian

    @classmethod
    def project(
            cls, samples: np.ndarray, log_weight_list: Optional[np.ndarray] = None, id_=None
    ) -> "AbstractMessage":
        """Calculates the sufficient statistics of a set of samples
        and returns the distribution with the appropriate parameters
        that match the sufficient statistics
        """
        # if weight_list aren't passed then equally weight all samples

        # Numerically stable weighting for very small/large weight_list

        # rescale coordinates to 'natural parameter space'
        if log_weight_list is None:
            log_weight_list = np.zeros_like(samples)

        log_w_max = np.max(log_weight_list, axis=0, keepdims=True)
        w = np.exp(log_weight_list - log_w_max)
        norm = w.mean(0)
        log_norm = np.log(norm) + log_w_max[0]

        tx = cls.to_canonical_form(samples)
        w /= norm
        suff_stats = (tx * w[None, ...]).mean(1)

        assert np.isfinite(suff_stats).all()

        cls_ = cls._projection_class or cls
        return cls_.from_sufficient_statistics(suff_stats, log_norm=log_norm, id_=id_)

    @classmethod
    def from_mode(
            cls, mode: np.ndarray, covariance: np.ndarray, id_
    ) -> "AbstractMessage":
        pass

    def log_normalisation(self, *elems: Union["AbstractMessage", float]) -> np.ndarray:
        """
        Calculates the log of the integral of the product of a
        set of distributions

        NOTE: ignores log normalisation
        """
        # Remove floats from messages passed
        dists: List[AbstractMessage] = [
            dist
            for dist in self._iter_dists(elems)
            if isinstance(dist, AbstractMessage)
        ]

        # Calculate log product of message normalisation
        log_norm = self.log_base_measure - self.log_partition
        log_norm += sum(dist.log_base_measure - dist.log_partition for dist in dists)

        # Calculate log normalisation of product of messages
        prod_dist = self.sum_natural_parameters(*dists)
        log_norm -= prod_dist.log_base_measure - prod_dist.log_partition

        return log_norm

    @staticmethod
    def _iter_dists(dists) -> Iterator[Union["AbstractMessage", float]]:
        for elem in dists:
            if isinstance(elem, Prior):
                yield elem
            elif np.isscalar(elem):
                yield elem
            else:
                for dist in elem:
                    yield dist

    def update_invalid(self, other: "AbstractMessage") -> "AbstractMessage":
        valid = self.check_valid()
        if self.ndim:
            valid_parameters: Iterator[np.ndarray] = (
                np.where(valid, p, p_safe) for p, p_safe in zip(self, other)
            )
        else:
            # TODO: Fairly certain this would not work
            valid_parameters = iter(self if valid else other)
        return type(self)(*valid_parameters, log_norm=self.log_norm, id_=self.id)

    def check_support(self) -> np.ndarray:
        if self._parameter_support is not None:
            return reduce(
                and_,
                (
                    (p >= support[0]) & (p <= support[1])
                    for p, support in zip(self.parameters, self._parameter_support)
                ),
            )
        elif self.ndim:
            return np.array(True, dtype=bool, ndmin=self.ndim)
        return np.array([True])

    def check_finite(self) -> np.ndarray:
        return np.isfinite(self.natural_parameters).all(0)

    def check_valid(self) -> np.ndarray:
        return self.check_finite() & self.check_support()

    @cached_property
    def is_valid(self) -> Union[np.ndarray, np.bool_]:
        return np.all(self.check_finite()) and np.all(self.check_support())

    @staticmethod
    def _get_mean_variance(
            mean: np.ndarray, covariance: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        mean, covariance = np.asanyarray(mean), np.asanyarray(covariance)

        if not covariance.shape:
            # If variance is float simply pass through
            variance = covariance * np.ones_like(mean)
            if not variance.shape:
                variance = variance.item()
        elif mean.shape == covariance.shape:
            variance = np.asanyarray(covariance)
        elif covariance.shape == mean.shape * 2:
            # If 2D covariance matrix passed get diagonal
            inds = tuple(np.indices(mean.shape))
            variance = np.asanyarray(covariance)[inds * 2]
        else:
            raise ValueError(
                f"shape of covariance {covariance.shape} is invalid "
                f"must be (), {mean.shape}, or {mean.shape * 2}"
            )
        return mean, variance

    def __call__(self, x):
        return np.sum(self.logpdf(x))

    def factor_jacobian(
            self, x: np.ndarray, _variables: Optional[Tuple[str]] = ("x",)
    ) -> Union[np.ndarray, Tuple[np.ndarray, Tuple[np.ndarray, ...]]]:
        loglike, g = self.logpdf_gradient(x)
        g = np.expand_dims(g, list(range(loglike.ndim)))
        return loglike.sum(), (g,)

    def as_factor(self, variable: "Variable", name: Optional[str] = None):
        from autofit.graphical.factor_graphs import Factor

        if name is None:
            shape = self.shape
            clsname = type(self).__name__
            family = clsname[:-7] if clsname.endswith("Message") else clsname
            name = f"{family}Likelihood" + (str(shape) if shape else "")

        return Factor(
            self.logpdf,
            variable,
            name=name,
            factor_jacobian=self.factor_jacobian,
            plates=variable.plates,
            arg_names=["x"],
        )

    @classmethod
    def transformed(
            cls,
            transform: Union[AbstractDensityTransform, Type[AbstractDensityTransform]],
            clsname: Optional[str] = None,
            support: Optional[Tuple[Tuple[float, float], ...]] = None,
            wrapper_cls=None,
    ):
        # noinspection PyUnresolvedReferences
        """
        transforms the distribution according the passed transform,
        returns a newly created class that encodes the transformation.

        Parameters
        ----------
        wrapper_cls
        transform: AbstractDensityTransform
            object that transforms the density
        clsname: str, optional
            the class name of the newly created class.
            defaults to "Transformed<OriginalClassName>"
        support: Tuple[Tuple[float, float], optional
            the support of the new class. Generally this can be
            automatically calculated from the parent class

        Examples
        --------
        >>> from autofit.messages.normal import NormalMessage

        Normal distributions have infinite univariate support
        >>> NormalMessage._support
        ((-inf, inf),)

        We can tranform the NormalMessage to the unit interval
        using `transform.phi_transform`
        >>> UnitNormal = NormalMessage.transformed(transform.phi_transform)
        >>> message = UnitNormal(1.2, 0.8)
        >>> message._support
        ((0.0, 1.0),)

        Samples from the UnitNormal will exist in the Unit interval
        >>> samples = message.sample(1000)
        >>> samples.min(), samples.mean(), samples.max()
        (0.06631750944045942, 0.8183189295040845, 0.9999056316923468)

        Projections still work for the transformed class
        >>> UnitNormal.project(samples, samples*0)
        TransformedNormalMessage(mu=1.20273342, sigma=0.80929032)

        Can specify the name of the new transformed class
        >>> NormalMessage.transformed(transform.phi_transform, 'UnitNormal')(0, 1.)
        UnitNormal(mu=0, sigma=1.)

        The transformed objects are pickleable
        >>> import pickle
        >>> pickle.loads(pickle.dumps(message))
        TransformedNormalMessage(mu=1.2, sigma=0.8)

        The transformed objects also are normalised,
        >>> from scipy.integrate import quad
        >>> # noinspection PyTypeChecker
        >>> quad(message.pdf, 0, 1)
        (1.0000000000114622, 3.977073226302252e-09)

        Can also nest transforms
        >>> WeirdNormal = NormalMessage.transformed(
            transform.log_transform).transformed(
            transform.exp_transform)
        This transformation is equivalent to the identity transform!
        >>> WeirdNormal.project(NormalMessage(0.3, 0.8).sample(1000))
        Transformed2NormalMessage(mu=0.31663248, sigma=0.79426984)

        This functionality is more useful for applying linear shifts
        e.g.
        >>> ShiftedUnitNormal = NormalMessage.transformed(
            transform.phi_transform
        ).shifted(shift=0.7, scale=2.3)
        >>> ShiftedUnitNormal._support
        ((0.7, 3.0),)
        >>> samples = ShiftedUnitNormal(0.2, 0.8).sample(1000)
        >>> samples.min(), samples.mean(), samples.max()
        """
        from .transform_wrapper import TransformedWrapper

        wrapper_cls = wrapper_cls or TransformedWrapper
        return wrapper_cls(
            cls=cls,
            transform=transform,
            clsname=clsname,
            support=support,
        )

    @classmethod
    def shifted(
            cls,
            shift: float = 0,
            scale: float = 1,
            wrapper_cls=None,
    ):
        return cls.transformed(
            LinearShiftTransform(shift=shift, scale=scale),
            clsname=f"Shifted{cls.__name__}",
            wrapper_cls=wrapper_cls,
        )

    @classmethod
    def _reconstruct(
            cls,
            parameters: Tuple[np.ndarray, ...],
            log_norm: float,
            id_,
            lower_limit,
            upper_limit,
            *args,
    ):
        return cls(
            *parameters,
            log_norm=log_norm,
            id_=id_,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
        )

    def __reduce__(self):
        # serialises TransformedMessage during pickling
        return (
            self._reconstruct,
            (
                self.parameters,
                self.log_norm,
                self.id,
                self.lower_limit,
                self.upper_limit,
            ),
        )

    def _sample(self, n_samples):
        # Needed for nested TransformedMessage method resolution
        return self.sample(n_samples)

    @classmethod
    def _logpdf_gradient(cls, self, x):
        # Needed for nested TransformedMessage method resolution
        return cls.logpdf_gradient(self, x)

    @classmethod
    def _logpdf_gradient_hessian(cls, self, x):
        # Needed for nested TransformedMessage method resolution
        return cls.logpdf_gradient_hessian(self, x)


def map_dists(
        dists: Dict[str, AbstractMessage],
        values: Dict[str, np.ndarray],
        _call: str = "logpdf",
) -> Iterator[Tuple[str, np.ndarray]]:
    """
    Calls a method (default: logpdf) for each Message in dists
    on the corresponding value in values
    """
    for v in dists.keys() & values.keys():
        dist = dists[v]
        if isinstance(dist, AbstractMessage):
            yield v, getattr(dist, _call)(values[v])



import numpy as np
import pytest
from scipy import integrate, stats

import autofit.messages.normal
from autofit import graphical as graph


@pytest.fixture(name="q_cavity")
def make_q_cavity():
    return autofit.messages.normal.NormalMessage(-0.5, 0.5)


def test_integration(q_cavity, probit_factor):
    x = np.linspace(-3, 3, 2 ** 10)

    probit = stats.norm(loc=0.0, scale=1.0).cdf(x)
    q = stats.norm(loc=q_cavity.mean, scale=q_cavity.sigma).pdf(x)
    tilted_distribution = probit * q

    assert tilted_distribution.shape == (2 ** 10,)

    ni_0, ni_1, ni_2 = (
        integrate.trapz(x ** i * tilted_distribution, x) for i in range(3)
    )

    q_numerical = autofit.messages.normal.NormalMessage.from_sufficient_statistics(
        [ni_1 / ni_0, ni_2 / ni_0]
    )

    assert q_numerical.mean == pytest.approx(-0.253, rel=0.01)
    assert q_numerical.sigma == pytest.approx(0.462, rel=0.01)


def test_laplace_method(probit_factor, q_cavity, x):
    mf = graph.MeanField({x: q_cavity})
    probit_approx = graph.FactorApproximation(
        factor=probit_factor,
        cavity_dist=mf,
        factor_dist=mf,
        model_dist=mf,
    )

    opt = graph.LaplaceOptimiser()
    new_dist, s = opt.optimise_approx(probit_approx)
    q_probit_laplace = new_dist[x]

    # opt_probit = graph.OptFactor.from_approx(probit_approx)
    # result = opt_probit.maximise({x: 0.0})

    # q_probit_laplace = autofit.messages.normal.NormalMessage.from_mode(
    #     result.mode[x], covariance=result.hess_inv[x]
    # )

    assert q_probit_laplace.mean == pytest.approx(-0.258, rel=0.01)
    assert q_probit_laplace.sigma == pytest.approx(0.462, rel=0.01)



import numpy as np
# TODO: Use autofit class?
from scipy import stats

import autofit as af


def _gaussian(x, centre, normalization, sigma):
    return Gaussian(centre=centre, normalization=normalization, sigma=sigma)(x)


_norm = stats.norm(loc=0, scale=1.0)


# TODO: use autofit likelihood
def _likelihood(z, y):
    return np.multiply(-0.5, np.square(np.subtract(z, y)))


class Profile:
    def __init__(self, centre=0.0, normalization=0.01):
        """Represents an Abstract 1D profile.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the profile.
        """
        self.centre = centre
        self.normalization = normalization


class Gaussian(Profile):
    def __init__(
            self,
            centre=0.0,  # <- PyAutoFit recognises these constructor arguments
            normalization=0.1,  # <- are the Gaussian's model parameters.
            sigma=0.01,
    ):
        """Represents a 1D Gaussian profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the Gaussian profile.
        sigma
            The sigma value controlling the size of the Gaussian.
        """
        super().__init__(centre=centre, normalization=normalization)
        self.sigma = sigma  # We still need to set sigma for the Gaussian, of course.

    def __call__(self, xvalues):
        """
        Calculate the normalization of the profile on a line of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues : np.ndarray
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return np.multiply(
            np.divide(self.normalization, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )


def make_data(gaussian, x):
    model_line = gaussian(xvalues=x)
    signal_to_noise_ratio = 25.0
    noise = np.random.normal(0.0, 1.0 / signal_to_noise_ratio, len(x))
    y = model_line + noise
    return y


class Analysis(af.Analysis):
    def __init__(self, x, y, sigma=.04):
        self.x = x
        self.y = y
        self.sigma = sigma

    def log_likelihood_function(self, instance: Gaussian) -> np.array:
        """
        This function takes an instance created by the PriorModel and computes the
        likelihood that it fits the data.
        """
        y_model = instance(self.x)
        return np.sum(_likelihood(y_model, self.y) / self.sigma ** 2)



from autofit.graphical.expectation_propagation import EPHistory
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.mapper.prior_model.collection import CollectionPriorModel


class EPResult:
    def __init__(
            self,
            ep_history: EPHistory,
            declarative_factor,
            updated_ep_mean_field: EPMeanField,
    ):
        """
        The result of an EP Optimisation including its history, a declarative
        representation of the optimised graph and the resultant EPMeanField
        which comprises factors and resultant approximations of variables.

        Parameters
        ----------
        ep_history
            A history of the optimisation
        declarative_factor: AbstractDeclarativeFactor
            A declarative representation of the factor being optimised
        updated_ep_mean_field
            An updated mean field; effectively the result of the optimisation
        """
        self.ep_history = ep_history
        self.declarative_factor = declarative_factor
        self.updated_ep_mean_field = updated_ep_mean_field

    @property
    def model(self) -> CollectionPriorModel:
        """
        A collection populated with messages representing the posteriors of
        the EP Optimisation. Each item in the collection represents a single
        factor in the optimisation.
        """
        collection = CollectionPriorModel({
            factor.name: factor.prior_model
            for factor
            in self.declarative_factor.model_factors
        })
        arguments = {
            prior: self.updated_ep_mean_field.mean_field[
                prior
            ]
            for prior
            in collection.priors
        }

        return collection.gaussian_prior_model_for_arguments(
            arguments
        )

    @property
    def instance(self):
        """
        The median instance taken from the updated model
        """
        return self.model.instance_from_prior_medians()



from autofit.database import query as q


class TestCombination:
    def test_simple(
            self,
            less_than,
            greater_than,
            simple_and
    ):
        assert (q.Q(
            "a",
            less_than
        ) & q.Q(
            "a",
            greater_than
        )).query == simple_and.query

    def test_second_level(
            self,
            less_than,
            greater_than,
            second_level
    ):
        first = q.Q("a", less_than)
        second = q.Q(
            'a',
            q.Q('b', greater_than)
        )

        assert (first & second).query == second_level.query

    def test_complicated(
            self,
            less_than,
            greater_than
    ):
        first = q.Q(
            "a",
            q.Q(
                "b",
                q.And(
                    q.Q(
                        "c",
                        less_than
                    ),
                    greater_than
                )
            )
        )

        second = q.Q(
            "a",
            q.Q(
                "b",
                q.Q(
                    "c",
                    greater_than
                )
            )
        )

        combined = q.Q(
            "a",
            q.Q(
                "b",
                q.And(
                    q.Q(
                        "c",
                        q.And(
                            less_than,
                            greater_than
                        )
                    ),
                    greater_than
                )
            )
        )

        assert (first & second).query == combined.query



import pytest

import autofit as af


@pytest.fixture(name="factor")
def make_factor(hierarchical_factor):
    return hierarchical_factor.factors[0]


def test_optimise(factor):
    optimizer = af.DynestyStatic(maxcall=100)

    _, status = optimizer.optimise(factor, factor.mean_field_approximation())
    assert status


def test_instance(factor):
    prior_model = factor.prior_model
    assert len(prior_model.priors) == 3
    assert factor.log_likelihood_function(
        prior_model.instance_from_prior_medians()
    ) == pytest.approx(-3.2215236261987186, 1e-10)



import pytest

from autofit import database as db


@pytest.fixture(
    name="fit"
)
def make_fit():
    return db.Fit(
        id="id",
        info={
            "key": "value"
        }
    )


def test_create(
        fit
):
    assert fit.info["key"] == "value"


@pytest.fixture(
    autouse=True
)
def add_to_session(
        session,
        fit
):
    session.add(fit)
    session.commit()


def test_query(
        aggregator,
        fit
):
    assert aggregator.query(
        aggregator.info["key"] == "value"
    ) == [fit]



import numpy as np
import pytest

import autofit as af
from test_autofit.graphical.gaussian.model import Gaussian, make_data


@pytest.fixture(
    name="x"
)
def make_x():
    return np.arange(100)


@pytest.fixture(
    name="y"
)
def make_y(x):
    return make_data(Gaussian(centre=50.0, normalization=25.0, sigma=10.0), x)


@pytest.fixture(
    name="prior_model"
)
def make_prior_model():
    return af.PriorModel(
        Gaussian,
        centre=af.GaussianPrior(mean=50, sigma=20),
        normalization=af.GaussianPrior(mean=25, sigma=10),
        sigma=af.GaussianPrior(mean=10, sigma=10),
    )



import pytest

import autofit as af
from autofit import Result
from autofit.non_linear.grid.grid_search import ResultBuilder
from autofit.non_linear.grid.grid_search.job import JobResult
from autofit.non_linear.grid.grid_search.result_builder import Placeholder

from autofit.mock.mock import MockSearch, MockResult

@pytest.fixture(
    name="result_builder"
)
def make_result_builder():
    return ResultBuilder(
        lists=[[1], [2], [3]],
        grid_priors=[]
    )


@pytest.fixture(
    name="add_results"
)
def make_add_results(
        result_builder
):
    def add_results(*numbers):
        for number in numbers:
            result_builder.add(
                JobResult(
                    MockResult(
                        search=MockSearch(
                            name=str(number)
                        ),
                        model=af.Model(
                            af.Gaussian
                        )
                    ),
                    [],
                    number
                )
            )

    return add_results


def test_ordering(
        result_builder,
        add_results
):
    add_results(0, 1, 2)
    result = result_builder()
    assert [
               result.search.name
               for result in result.results
           ] == ['0', '1', '2']


@pytest.mark.parametrize(
    "numbers, t1, t2, t3",
    [
        (tuple(), Placeholder, Placeholder, Placeholder,),
        ((0,), Result, Placeholder, Placeholder,),
        ((1,), Placeholder, Result, Placeholder,),
        ((2,), Placeholder, Placeholder, Result,),
        ((1, 2,), Placeholder, Result, Result,),
    ]
)
def test_gaps(
        result_builder,
        add_results,
        numbers,
        t1, t2, t3
):
    add_results(*numbers)
    result = result_builder()

    first, second, third = result.results
    assert isinstance(first, t1)
    assert isinstance(second, t2)
    assert isinstance(third, t3)


def test_log_likelihoods(
        result_builder
):
    assert (result_builder().log_likelihoods_native == [
        None, None, None
    ]).all()


def test_best_result(
        result_builder,
        add_results
):
    add_results(1)
    assert isinstance(
        result_builder().best_result,
        Result
    )


def test_placeholder_samples():
    assert Placeholder().samples.anything is None



import pickle
import shutil
from os import path

import numpy as np
import pytest

import autofit as af
from autoconf import conf

from autofit.mock.mock_model import MockClassx4, MockClassx2Tuple
from autofit.mock.mock import MockSearch, MockSamples

pytestmark = pytest.mark.filterwarnings("ignore::FutureWarning")


@pytest.fixture(name="mapper")
def make_mapper():
    return af.ModelMapper()


@pytest.fixture(name="mock_list")
def make_mock_list():
    return [af.PriorModel(MockClassx4), af.PriorModel(MockClassx4)]


@pytest.fixture(name="result")
def make_result():
    mapper = af.ModelMapper()
    mapper.component = MockClassx2Tuple
    # noinspection PyTypeChecker
    return af.Result(
        samples=MockSamples(gaussian_tuples=[(0, 0), (1, 0)]),
        model=mapper,
        search=MockSearch(),
    )


class TestResult:
    def test_model(self, result):
        component = result.model.component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 0.2
        assert component.one_tuple.one_tuple_1.sigma == 0.2

    def test_model_absolute(self, result):
        component = result.model_absolute(a=2.0).component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 2.0
        assert component.one_tuple.one_tuple_1.sigma == 2.0

    def test_model_relative(self, result):
        component = result.model_relative(r=1.0).component
        assert component.one_tuple.one_tuple_0.mean == 0
        assert component.one_tuple.one_tuple_1.mean == 1
        assert component.one_tuple.one_tuple_0.sigma == 0.0
        assert component.one_tuple.one_tuple_1.sigma == 1.0

    def test_raises(self, result):
        with pytest.raises(af.exc.PriorException):
            result.model.mapper_from_gaussian_tuples(
                result.samples._gaussian_tuples, a=2.0, r=1.0
            )


class TestLabels:
    def test_param_names(self):
        model = af.PriorModel(MockClassx4)
        assert [
                   "one",
                   "two",
                   "three",
                   "four",
               ] == model.model_component_and_parameter_names

    def test_label_config(self):
        assert conf.instance["notation"]["label"]["label"]["one"] == "one_label"
        assert conf.instance["notation"]["label"]["label"]["two"] == "two_label"
        assert conf.instance["notation"]["label"]["label"]["three"] == "three_label"
        assert conf.instance["notation"]["label"]["label"]["four"] == "four_label"


test_path = path.join(
    "{}".format(path.dirname(path.realpath(__file__))), "files", "phase"
)


class TestMovePickleFiles:
    def test__move_pickle_files(self):

        search = MockSearch()
        search.paths = af.DirectoryPaths(
            name="pickles",
            path_prefix=path.join("non_linear", "abstract_search")
        )

        pickle_paths = [
            path.join(
                conf.instance.output_path, "non_linear", "abstract_search", "pickles"
            )
        ]

        arr = np.ones((3, 3))

        with open(path.join(pickle_paths[0], "test.pickle"), "wb") as f:
            pickle.dump(arr, f)

        pickle_paths = [
            path.join(
                conf.instance.output_path,
                "non_linear",
                "abstract_search",
                "pickles",
                "test.pickle",
            )
        ]

        search.paths._move_pickle_files(pickle_files=pickle_paths)

        with open(path.join(pickle_paths[0]), "rb") as f:
            arr_load = pickle.load(f)

        assert (arr == arr_load).all()

        if path.exists(test_path):
            shutil.rmtree(test_path)



import math
from typing import List, Optional

import numpy as np

from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples.sample import Sample, load_from_table
from .samples import Samples


class PDFSamples(Samples):
    def __init__(
            self,
            model: AbstractPriorModel,
            sample_list: List[Sample],
            unconverged_sample_size: int = 100,
            time: Optional[float] = None,
            results_internal: Optional = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results_internal of
        a `NonLinearSearch` (e.g. as files on your hard-disk) and Python.

        For example, the output class can be used to load an instance of the best-fit model, get an instance of any
        individual sample by the `NonLinearSearch` and return information on the likelihoods, errors, etc.

        This class stores samples of searches which provide the probability distribution function (PDF) of the
        model fit (e.g. nested samplers, MCMC).

        To use a library's in-built visualization tools results are optionally stored in their native internal format
        using the `results_internal` attribute.

        Parameters
        ----------
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        sample_list
            The list of `Samples` which contains the paramoeters, likelihood, weights, etc. of every sample taken
            by the non-linear search.
        unconverged_sample_size
            If the samples are for a search that is yet to convergence, a reduced set of samples are used to provide
            a rough estimate of the parameters. The number of samples is set by this parameter.
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        results_internal
            The nested sampler's results in their native internal format for interfacing its visualization library.
        """

        super().__init__(
            model=model,
            sample_list=sample_list,
            time=time,
            results_internal=results_internal
        )

        self._unconverged_sample_size = int(unconverged_sample_size)

    @classmethod
    def from_table(cls, filename: str, model, number_live_points=None):
        """
        Write a table of parameters, posteriors, priors and likelihoods

        Parameters
        ----------
        filename
            Where the table is to be written
        """
        from .stored import StoredSamples

        sample_list = load_from_table(
            filename=filename
        )

        return StoredSamples(
            model=model,
            sample_list=sample_list
        )

    @property
    def unconverged_sample_size(self):
        """
        If a set of samples are unconverged, alternative methods to compute their means, errors, etc are used as
        an alternative to corner.py.

        These use a subset of samples spanning the range from the most recent sample to the valaue of the
        unconverted_sample_size. However, if there are fewer samples than this size, we change the size to be the
        the size of the total number of samples.
        """
        if self.total_samples > self._unconverged_sample_size:
            return self._unconverged_sample_size
        return self.total_samples

    @property
    def pdf_converged(self) -> bool:
        """
        To analyse and visualize samples the analysis must be sufficiently converged to produce smooth enough
        PDF for error estimate and PDF generation.

        This property checks whether the non-linear search's samples are sufficiently converged for this, by checking
        if one sample's weight contains > 99% of the weight. If this is the case, it implies the convergence necessary
        for error estimate and visualization has not been met.

        This does not necessarily imply the `NonLinearSearch` has converged overall, only that errors and visualization
        can be performed numerically.
        """
        if np.max(self.weight_list) > 0.99:
            return False
        return True

    @property
    def median_pdf_vector(self) -> [float]:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a list of values.
        """
        if self.pdf_converged:
            return [
                quantile(x=params, q=0.5, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]
        return self.max_log_likelihood_vector

    @property
    def median_pdf_instance(self) -> ModelInstance:
        """
        The median of the probability density function (PDF) of every parameter marginalized in 1D, returned
        as a model instance.
        """
        return self.model.instance_from_vector(vector=self.median_pdf_vector)

    def vector_at_sigma(self, sigma: float) -> [(float, float)]:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as two lists of values corresponding to the lower and upper values parameter values.

        For example, if sigma is 1.0, the marginalized values of every parameter at 31.7% and 68.2% percentiles of each
        PDF is returned.

        This does not account for covariance between parameters. For example, if two parameters (x, y) are degenerate
        whereby x decreases as y gets larger to give the same PDF, this function will still return both at their
        upper values. Thus, caution is advised when using the function to reperform a model-fits.

        For Dynesty, this is estimated using corner.py if the samples have converged, by sampling the density
        function at an input PDF %. If not converged, a crude estimate using the range of values of the current
        physical live points is used.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """

        if self.pdf_converged:

            low_limit = (1 - math.erf(sigma / math.sqrt(2))) / 2

            lower_errors = [
                quantile(x=params, q=low_limit, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]
            upper_errors = [
                quantile(x=params, q=1 - low_limit, weights=self.weight_list)[0]
                for params in self.parameters_extract
            ]

            return [(lower, upper) for lower, upper in zip(lower_errors, upper_errors)]

        parameters_min = list(
            np.min(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )
        parameters_max = list(
            np.max(self.parameter_lists[-self.unconverged_sample_size:], axis=0)
        )

        return [
            (parameters_min[index], parameters_max[index])
            for index in range(len(parameters_min))
        ]

    def vector_at_upper_sigma(self, sigma: float) -> [float]:
        """
        The upper value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(map(lambda param: param[1], self.vector_at_sigma(sigma)))

    def vector_at_lower_sigma(self, sigma: float) -> [float]:
        """
        The lower value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma limit within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return list(map(lambda param: param[0], self.vector_at_sigma(sigma)))

    def instance_at_sigma(self, sigma: float) -> ModelInstance:
        """
        The value of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as a list of model instances corresponding to the lower and upper values estimated from the PDF.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.vector_at_sigma(sigma=sigma),
        )

    def instance_at_upper_sigma(self, sigma: float) -> ModelInstance:
        """
        The upper value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a model instance.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.vector_at_upper_sigma(sigma=sigma),
        )

    def instance_at_lower_sigma(self, sigma: float) -> ModelInstance:
        """
        The lower value of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a model instance.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.vector_at_lower_sigma(sigma=sigma),
        )

    def error_vector_at_sigma(self, sigma: float) -> [(float, float)]:
        """
        The lower and upper error of every parameter marginalized in 1D at an input sigma value of its probability
        density function (PDF), returned as a list.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        error_vector_lower = self.error_vector_at_lower_sigma(sigma=sigma)
        error_vector_upper = self.error_vector_at_upper_sigma(sigma=sigma)
        return [(lower, upper) for lower, upper in zip(error_vector_lower, error_vector_upper)]

    def error_vector_at_upper_sigma(self, sigma: float) -> [float]:
        """
        The upper error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        uppers = self.vector_at_upper_sigma(sigma=sigma)
        return list(
            map(
                lambda upper, median_pdf: upper - median_pdf,
                uppers,
                self.median_pdf_vector,
            )
        )

    def error_vector_at_lower_sigma(self, sigma: float) -> [float]:
        """
        The lower error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a list.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        lowers = self.vector_at_lower_sigma(sigma=sigma)
        return list(
            map(
                lambda lower, median_pdf: median_pdf - lower,
                lowers,
                self.median_pdf_vector,
            )
        )

    def error_magnitude_vector_at_sigma(self, sigma: float) -> [float]:
        """
        The magnitude of every error after marginalization in 1D at an input sigma value of the probability density
        function (PDF), returned as two lists of values corresponding to the lower and upper errors.

        For example, if sigma is 1.0, the difference in the inferred values marginalized at 31.7% and 68.2% percentiles
        of each PDF is returned.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        uppers = self.vector_at_upper_sigma(sigma=sigma)
        lowers = self.vector_at_lower_sigma(sigma=sigma)
        return list(map(lambda upper, lower: upper - lower, uppers, lowers))

    def error_instance_at_sigma(self, sigma: float) -> ModelInstance:
        """
        The error of every parameter marginalized in 1D at an input sigma value of its probability density function
        (PDF), returned as a list of model instances corresponding to the lower and upper errors.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        ----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.error_magnitude_vector_at_sigma(sigma=sigma),
        )

    def error_instance_at_upper_sigma(self, sigma: float) -> ModelInstance:
        """
        The upper error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a model instance.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.error_vector_at_upper_sigma(sigma=sigma),
        )

    def error_instance_at_lower_sigma(self, sigma: float) -> ModelInstance:
        """
        The lower error of every parameter marginalized in 1D at an input sigma value of its probability density
        function (PDF), returned as a model instance.

        See vector_at_sigma for a full description of how the parameters at sigma are computed.

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """
        return self.model.instance_from_vector(
            vector=self.error_vector_at_lower_sigma(sigma=sigma),
        )

    def gaussian_priors_at_sigma(self, sigma: float) -> [List]:
        """
        `GaussianPrior`s of every parameter used to link its inferred values and errors to priors used to sample the
        same (or similar) parameters in a subsequent search, where:

        - The mean is given by their most-probable values (using median_pdf_vector).
        - Their errors are computed at an input sigma value (using errors_at_sigma).

        Parameters
        -----------
        sigma
            The sigma within which the PDF is used to estimate errors (e.g. sigma = 1.0 uses 0.6826 of the PDF).
        """

        means = self.median_pdf_vector

        uppers = self.vector_at_upper_sigma(sigma=sigma)
        lowers = self.vector_at_lower_sigma(sigma=sigma)

        # noinspection PyArgumentList
        sigmas = list(
            map(
                lambda mean, upper, lower: max([upper - mean, mean - lower]),
                means,
                uppers,
                lowers,
            )
        )

        return list(map(lambda mean, sigma: (mean, sigma), means, sigmas))

    def log_likelihood_from_sample_index(self, sample_index: int) -> float:
        """
        The log likelihood of an individual sample of the non-linear search.

        Parameters
        ----------
        sample_index
            The index of the sample in the non-linear search, e.g. 0 gives the first sample.
        """
        raise NotImplementedError()

    def vector_drawn_randomly_from_pdf(self) -> [float]:
        """
        The parameter vector of an individual sample of the non-linear search drawn randomly from the PDF, returned as
        a 1D list.

        The draw is weighted by the sample weights to ensure that the sample is drawn from the PDF (which is important
        for non-linear searches like nested sampling).
        """
        sample_index = np.random.choice(a=range(len(self.sample_list)), p=self.weight_list)

        return self.parameter_lists[sample_index][:]

    def instance_drawn_randomly_from_pdf(self) -> ModelInstance:
        """
        The parameter instance of an individual sample of the non-linear search drawn randomly from the PDF, returned
        as a 1D list.

        The draw is weighted by the sample weights to ensure that the sample is drawn from the PDF (which is important
        for non-linear searches like nested sampling).
        """
        return self.model.instance_from_vector(vector=self.vector_drawn_randomly_from_pdf())

    def vector_from_sample_index(self, sample_index: int) -> [float]:
        """
        The parameters of an individual sample of the non-linear search, returned as a 1D list.

        Parameters
        ----------
        sample_index
            The index of the sample in the non-linear search, e.g. 0 gives the first sample.
        """
        raise NotImplementedError()

    def offset_vector_from_input_vector(self, input_vector: List) -> [float]:
        """
        The values of an input_vector offset by the median_pdf_vector (the PDF medians).

        If the 'true' values of a model are known and input as the input_vector, this function returns the results
        of the `NonLinearSearch` as values offset from the 'true' model. For example, a value 0.0 means the non-linear
        search estimated the model parameter value correctly.

        Parameters
        ----------
        input_vector
            A 1D list of values the most-probable model is offset by. This list must have the same dimensions as free
            parameters in the model-fit.
        """
        return list(
            map(
                lambda input, median_pdf: median_pdf - input,
                input_vector,
                self.median_pdf_vector,
            )
        )

    def covariance_matrix(self) -> np.ndarray:
        """
        Compute the covariance matrix of the non-linear search samples, using the method `np.cov()` which is described
        at the following link:

        https://numpy.org/doc/stable/reference/generated/numpy.cov.html

        Follow that link for a description of what the covariance matrix is.

        Returns
        -------
        ndarray
            A covariance matrix of shape [total_parameters, total_parameters] for the model parameters of the
            non-linear search.
        """
        return np.cov(m=self.parameter_lists, rowvar=False, aweights=self.weight_list)


def quantile(x, q, weights=None):
    """
    Copied from corner.py

    Compute sample quantiles with support for weighted samples.

    Note
    ----
    When ``weight_list`` is ``None``, this method simply calls numpy's percentile
    function with the values of ``q`` multiplied by 100.

    Parameters
    ----------
    x : array_like[nsamples,]
       The samples.
    q : array_like[nquantiles,]
       The list of quantiles to compute. These should all be in the range
       ``[0, 1]``.
    weights: Optional[array_like[nsamples,]]
        An optional weight corresponding to each sample. These

    Returns
    -------
    quantiles : array_like[nquantiles,]
        The sample quantiles computed at ``q``.

    Raises
    ------
    ValueError
        For invalid quantiles; ``q`` not in ``[0, 1]`` or dimension mismatch between ``x`` and ``weight_list``.
    """
    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    if weights is None:
        return np.percentile(x, list(100.0 * q))
    else:
        weights = np.atleast_1d(weights)
        if len(x) != len(weights):
            raise ValueError("Dimension mismatch: len(weight_list) != len(x)")
        idx = np.argsort(x)
        sw = weights[idx]
        cdf = np.cumsum(sw)[:-1]
        cdf /= cdf[-1]
        cdf = np.append(0, cdf)
        return np.interp(q, cdf, x[idx]).tolist()



from pathlib import Path

import autofit as af
from autofit.non_linear.paths.sub_directory_paths import (
    SubDirectoryPaths, SubDirectoryPathsDirectory,
    SubDirectoryPathsDatabase
)


def test_directory():
    paths = af.DirectoryPaths()
    subdirectory_path = SubDirectoryPaths(
        parent=paths,
        analysis_name="name"
    )
    assert isinstance(
        subdirectory_path,
        SubDirectoryPathsDirectory
    )
    assert subdirectory_path.output_path == str(Path(
        paths.output_path
    ) / "name")


def test_database(session):
    paths = af.DatabasePaths(session)
    subdirectory_path = SubDirectoryPaths(
        parent=paths,
        analysis_name="name"
    )
    assert isinstance(
        subdirectory_path,
        SubDirectoryPathsDatabase
    )
    assert subdirectory_path.output_path == str(Path(
        paths.output_path
    ) / "name")


def test_is_flat():
    paths = af.DirectoryPaths()
    subdirectory_path = SubDirectoryPaths(
        parent=paths,
        analysis_name="name",
        is_flat=True
    )
    assert subdirectory_path.parent is paths

    subdirectory_path = SubDirectoryPaths(
        parent=subdirectory_path,
        analysis_name="name",
        is_flat=True,
    )
    assert subdirectory_path.parent is paths



def test_step_id(
        step_1,
        step_2
):
    assert isinstance(
        step_1.id,
        str
    )

    assert step_1 == step_1.id
    assert step_1 == step_1

    assert step_2 != step_1
    assert step_2 == step_2


def test_revision_steps(
        step_1,
        step_2,
        revision_1,
        revision_2
):
    assert revision_1.steps == (step_1,)
    assert revision_2.steps == (step_1, step_2)


def test_revision_ids(
        revision_1,
        revision_2
):
    assert revision_1.id != revision_2.id
    assert isinstance(
        revision_1.id,
        str
    )

    assert revision_1 != revision_2
    assert revision_1 == revision_1
    assert revision_1 == revision_1.id


def test_difference(
        revision_1,
        revision_2,
        step_2
):
    assert (revision_2 - revision_1).steps == (step_2,)


def test_get_steps(
        migrator,
        step_1,
        step_2,
        revision_1,
        revision_2
):
    assert migrator.get_steps() == (
        step_1, step_2
    )
    assert migrator.get_steps(
        revision_1.id
    ) == (step_2,)
    assert migrator.get_steps(
        revision_2.id
    ) == ()

    assert migrator.get_steps("random") == (
        step_1, step_2
    )



import pytest

import autofit as af
import autofit.graphical as g
from autoconf.conf import output_path_for_test
from autofit.graphical.utils import Status
from test_autofit.graphical.gaussian.model import Analysis


@pytest.fixture(name="analysis")
def make_analysis(x, y):
    return Analysis(x=x, y=y)


@pytest.fixture(name="factor_model")
def make_factor_model(prior_model, analysis):
    return g.AnalysisFactor(prior_model, analysis=analysis)


@pytest.fixture(name="laplace")
def make_laplace():
    return g.LaplaceOptimiser()


@pytest.fixture(name="dynesty")
def make_dynesty():
    return af.DynestyStatic(name="", maxcall=10)


def test_default(factor_model, laplace):
    model = factor_model.optimise(laplace)

    assert model.centre.mean == pytest.approx(50, rel=0.1)
    assert model.normalization.mean == pytest.approx(25, rel=0.1)
    assert model.sigma.mean == pytest.approx(10, rel=0.1)


def test_set_model_identifier(dynesty, prior_model, analysis):
    dynesty.fit(prior_model, analysis)

    identifier = dynesty.paths.identifier
    assert identifier is not None

    prior_model.centre = af.GaussianPrior(mean=20, sigma=20)
    dynesty.fit(prior_model, analysis)

    assert identifier != dynesty.paths.identifier


class TestDynesty:
    @output_path_for_test()
    def test_optimisation(self, factor_model, laplace, dynesty):
        factor_model.optimiser = dynesty
        factor_model.optimise(laplace)

    def test_null_paths(self, factor_model):
        optimizer = af.DynestyStatic(maxcall=10)
        result, status = optimizer.optimise(
            factor_model, factor_model.mean_field_approximation()
        )

        assert isinstance(result, g.EPMeanField)
        assert isinstance(status, Status)

    @output_path_for_test()
    def test_optimise(self, factor_model, dynesty):
        result, status = dynesty.optimise(
            factor_model, factor_model.mean_field_approximation()
        )

        assert isinstance(result, g.EPMeanField)
        assert isinstance(status, Status)
